#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from __future__ import annotations

import sys

import sklearn.metrics as _metrics
import numpy as np


def get_max_score_from_system_size() -> float:
    """
    recreates the call to sys.float_info.max in order to
    follow what was used in ads repo.

    Making into function which executes each time
    so unforseen issues are less likely when running on
    distributed env
    """

    return sys.float_info.max**0.5


def renumber_clusters(clusters: np.ndarray, reorder: bool):
    """Takes in cluster identifiers and renumbers them.
        After merging or reclustering there are many cluster numbers left blank and need to be renumbered
            Example: [0, 1, 2, 5, 7]
        Additionally the clusters are reordered from largest cluster to smallest

    Args:
        clusters (list|np.array): an array in which a cluster number is defined for each load shape

    Returns:
        clusters (np.array): an array in which a cluster number is defined for each load shape
    """

    if reorder:
        # if outlier cluster exists, don't include it in the ordering
        uniq_id, counts = np.unique(clusters[clusters != -1], return_counts=True)
        count_order = np.argsort(counts)[::-1]

        uniq_id = uniq_id[count_order]

    else:
        uniq_id = np.unique(clusters)

    # if outlier cluster exists, don't change it
    conv = {-1: -1}
    conv.update({uniq_id[i]: i for i in range(len(uniq_id))})

    clusters = np.array([conv[idx] for idx in clusters])

    return clusters


def merge_small_clusters(clusters: np.ndarray, min_cluster_size: int):
    """
    OG DOCSTRING:
    Merges clusters which consist of less than the minumum number into the outlier cluster

    Args:
        clusters (list|np.array): A list defining what cluster each load shape belongs to
        min_cluster_size (int): Minumum number of meters for a cluster
            Options: 2 < val

    Returns:
        _type_: _description_
    """

    uniq_ids, uniq_counts = np.unique(clusters, return_counts=True)

    uniq_counts = uniq_counts[uniq_ids != -1]
    uniq_ids = uniq_ids[uniq_ids != -1]

    outlier_ids = uniq_ids[uniq_counts < min_cluster_size]
    clusters[np.isin(clusters, outlier_ids)] = -1

    return renumber_clusters(clusters, reorder=True)


def score_clusters(
    data: np.ndarray,
    labels: np.ndarray,
    n_cluster_lower: int,
    score_choice="silhouette",
    dist_metric="euclidean",
    min_cluster_size=2,
    max_non_outlier_cluster_count=200,
) -> tuple[float, bool]:
    """
    ---
    Original docstring:

    Score clusters of the given data with the selected choices.
    Small clusters are first merged to only score clusters above the minimum size
    and not in the outlier cluster.

    Args:
        data (np.array): Load shapes being clustered
        labels (list|np.array): A list defining what cluster each load shape belongs to

    Returns:
        score (float): Lower is better
        unable_to_calc_score (bool): Boolean that if true, means max score was used
    """

    # merge clusters to outlier cluster
    labels = merge_small_clusters(labels, min_cluster_size)

    non_outlier_cluster_count = labels.max() + 1
    if non_outlier_cluster_count < n_cluster_lower:
        return get_max_score_from_system_size(), True

    if non_outlier_cluster_count > max_non_outlier_cluster_count:
        return get_max_score_from_system_size(), True

    # don't include outlier cluster in scoring
    idx = np.argwhere(labels != -1).flatten()
    data_non_outlier = data[idx, :]
    labels_non_outlier = labels[idx]

    score_error = False
    if score_choice == "silhouette":
        try:
            # if sample size is a number then it randomly samples within the group, None looks at all
            # if using silhouette score for large datasets, might want to specify sample_size
            score = float(
                _metrics.silhouette_score(
                    data_non_outlier,
                    labels_non_outlier,
                    metric=dist_metric,
                    sample_size=10_000,
                )
            )

        except Exception:
            score = float(10.0)
            score_error = True

    elif score_choice == "silhouette_median":
        try:
            # if this is too computationally intensive, could sample clusters instead
            score_all = -10 * _metrics.silhouette_samples(
                data_non_outlier, labels_non_outlier, metric=dist_metric
            )  # type: ignore
            score = float(np.median(score_all[idx]))
        except Exception:
            score = float(10.0)
            score_error = True

    elif score_choice in ["variance_ratio", "calinski-harabasz"]:
        try:
            score = -1 * _metrics.calinski_harabasz_score(
                data_non_outlier, labels_non_outlier
            )
        except Exception:
            score = get_max_score_from_system_size()
            score_error = True

    elif score_choice == "davies-bouldin":
        try:
            score = float(
                _metrics.davies_bouldin_score(data_non_outlier, labels_non_outlier)
            )
        except Exception:
            score = get_max_score_from_system_size()
            score_error = True

    else:
        raise ValueError(
            f"{score_choice} is not a recognized clustering scoring function"
        )

    return score, score_error



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from scipy.stats import t

from opendsm.eemeter.models.daily.model import DailyModel

__all__ = ("metered_savings", "modeled_savings")


def _compute_ols_error(
    t_stat,
    rmse_base_residuals,
    post_obs,
    base_obs,
    base_avg,
    post_avg,
    base_var,
    nprime,
):
    ols_model_agg_error = (
        (t_stat * rmse_base_residuals * post_obs)
        / (base_obs**0.5)
        * (1.0 + ((base_avg - post_avg) ** 2.0 / base_var)) ** 0.5
    )

    ols_noise_agg_error = (
        t_stat * rmse_base_residuals * (post_obs * base_obs / nprime) ** 0.5
    )

    ols_total_agg_error = (ols_model_agg_error**2.0 + ols_noise_agg_error**2.0) ** 0.5

    return ols_total_agg_error, ols_model_agg_error, ols_noise_agg_error


def _compute_fsu_error(
    t_stat,
    interval,
    post_obs,
    total_base_energy,
    rmse_base_residuals,
    base_avg,
    base_obs,
    nprime,
):
    if interval.startswith("billing"):
        a_coeff = -0.00022
        b_coeff = 0.03306
        c_coeff = 0.94054
        months_reporting = float(post_obs)
    else:  # daily
        a_coeff = -0.00024
        b_coeff = 0.03535
        c_coeff = 1.00286
        months_reporting = float(post_obs) / 30.0

    fsu_error_band = total_base_energy * (
        t_stat
        * (a_coeff * months_reporting**2.0 + b_coeff * months_reporting + c_coeff)
        * (rmse_base_residuals / base_avg)
        * ((base_obs / nprime) * (1.0 + (2.0 / nprime)) * (1.0 / post_obs)) ** 0.5
    )

    return fsu_error_band


def _compute_error_bands_metered_savings(
    totals_metrics, results, interval, confidence_level
):
    num_parameters = float(totals_metrics.num_parameters)

    base_obs = float(totals_metrics.observed_length)
    if (interval.startswith("billing")) & (len(results.dropna().index) > 0):
        post_obs = float(round((results.index[-1] - results.index[0]).days / 30.0))
    else:
        post_obs = float(results["reporting_observed"].dropna().shape[0])

    degrees_of_freedom = float(base_obs - num_parameters)
    single_tailed_confidence_level = 1 - ((1 - confidence_level) / 2)
    t_stat = t.ppf(single_tailed_confidence_level, degrees_of_freedom)

    rmse_base_residuals = float(totals_metrics.rmse_adj)
    autocorr_resid = totals_metrics.autocorr_resid

    base_avg = float(totals_metrics.observed_mean)
    post_avg = float(results["reporting_observed"].mean())

    base_var = float(totals_metrics.observed_variance)

    # these result in division by zero error for fsu_error_band
    if (
        post_obs == 0
        or autocorr_resid is None
        or abs(autocorr_resid) == 1
        or base_obs == 0
        or base_avg == 0
        or base_var == 0
    ):
        return None

    autocorr_resid = float(autocorr_resid)

    nprime = float(base_obs * (1 - autocorr_resid) / (1 + autocorr_resid))

    total_base_energy = float(base_avg * base_obs)

    ols_total_agg_error, ols_model_agg_error, ols_noise_agg_error = _compute_ols_error(
        t_stat,
        rmse_base_residuals,
        post_obs,
        base_obs,
        base_avg,
        post_avg,
        base_var,
        nprime,
    )

    fsu_error_band = _compute_fsu_error(
        t_stat,
        interval,
        post_obs,
        total_base_energy,
        rmse_base_residuals,
        base_avg,
        base_obs,
        nprime,
    )

    return {
        "FSU Error Band": fsu_error_band,
        "OLS Error Band": ols_total_agg_error,
        "OLS Error Band: Model Error": ols_model_agg_error,
        "OLS Error Band: Noise": ols_noise_agg_error,
    }


def metered_savings(
    baseline_model,
    reporting_meter_data,
    temperature_data,
    with_disaggregated=False,
    confidence_level=0.90,
    predict_kwargs=None,
    degc: bool = False,
    billing_data: bool = False,
):
    """Compute metered savings, i.e., savings in which the baseline model
    is used to calculate the modeled usage in the reporting period. This
    modeled usage is then compared to the actual usage from the reporting period.
    Also compute two measures of the uncertainty of the aggregate savings estimate,
    a fractional savings uncertainty (FSU) error band and an OLS error band. (To convert
    the FSU error band into FSU, divide by total estimated savings.)

    Parameters
    ----------
    baseline_model : :any:`eemeter.CalTRACKUsagePerDayModelResults`
        Object to use for predicting pre-intervention usage.
    reporting_meter_data : :any:`pandas.DataFrame`
        The observed reporting period data (totals). Savings will be computed for the
        periods supplied in the reporting period data.
    temperature_data : :any:`pandas.Series`
        Hourly-frequency timeseries of temperature data during the reporting
        period.
    with_disaggregated : :any:`bool`, optional
        If True, calculate baseline counterfactual disaggregated usage
        estimates. Savings cannot be disaggregated for metered savings. For
        that, use :any:`eemeter.modeled_savings`.
    confidence_level : :any:`float`, optional
        The two-tailed confidence level used to calculate the t-statistic used
        in calculation of the error bands.

        Ignored if not computing error bands.
    predict_kwargs : :any:`dict`, optional
        Extra kwargs to pass to the baseline_model.predict method.
    degc : :any 'bool'
        Relevant temperature units; defaults to False (i.e. Fahrenheit).

    Returns
    -------
    results : :any:`pandas.DataFrame`
        DataFrame with metered savings, indexed with
        ``reporting_meter_data.index``. Will include the following columns:

        - ``counterfactual_usage`` (baseline model projected into reporting period)
        - ``reporting_observed`` (given by reporting_meter_data)
        - ``metered_savings``

        If `with_disaggregated` is set to True, the following columns will also
        be in the results DataFrame:

        - ``counterfactual_base_load``
        - ``counterfactual_heating_load``
        - ``counterfactual_cooling_load``

    error_bands : :any:`dict`, optional
        If baseline_model is an instance of CalTRACKUsagePerDayModelResults,
        will also return a dictionary of FSU and OLS error bands for the
        aggregated energy savings over the post period.
    """
    if degc == True:
        temperature_data = 32 + (temperature_data * 1.8)

    if predict_kwargs is None:
        predict_kwargs = {}

    model_type = None
    if isinstance(baseline_model, DailyModel):
        raise NotImplementedError(
            "Use predict() with daily and billing models to compute metered savings."
        )

    prediction_index = reporting_meter_data.index
    model_prediction = baseline_model.predict(
        prediction_index, temperature_data, **predict_kwargs
    )

    predicted_baseline_usage = model_prediction.result

    # CalTrack 3.5.1
    counterfactual_usage = predicted_baseline_usage["predicted_usage"].to_frame(
        "counterfactual_usage"
    )

    reporting_observed = reporting_meter_data["value"].to_frame("reporting_observed")

    def metered_savings_func(row):
        return row.counterfactual_usage - row.reporting_observed

    results = reporting_observed.join(counterfactual_usage).assign(
        metered_savings=metered_savings_func
    )

    results = results.dropna().reindex(results.index)  # carry NaNs

    # compute t-statistic associated with n degrees of freedom
    # and a two-tailed confidence level.
    error_bands = None
    return results, error_bands


def _compute_error_bands_modeled_savings(
    totals_metrics_baseline,
    totals_metrics_reporting,
    results,
    interval_baseline,
    interval_reporting,
    confidence_level,
):
    num_parameters_baseline = float(totals_metrics_baseline.num_parameters)
    num_parameters_reporting = float(totals_metrics_reporting.num_parameters)

    base_obs_baseline = float(totals_metrics_baseline.observed_length)
    base_obs_reporting = float(totals_metrics_reporting.observed_length)

    if (interval_baseline.startswith("billing")) & (len(results.dropna().index) > 0):
        post_obs_baseline = float(
            round((results.index[-1] - results.index[0]).days / 30.0)
        )
    else:
        post_obs_baseline = float(results["modeled_baseline_usage"].dropna().shape[0])

    if (interval_reporting.startswith("billing")) & (len(results.dropna().index) > 0):
        post_obs_reporting = float(
            round((results.index[-1] - results.index[0]).days / 30.0)
        )
    else:
        post_obs_reporting = float(results["modeled_reporting_usage"].dropna().shape[0])

    degrees_of_freedom_baseline = float(base_obs_baseline - num_parameters_baseline)
    degrees_of_freedom_reporting = float(base_obs_reporting - num_parameters_reporting)
    single_tailed_confidence_level = 1 - ((1 - confidence_level) / 2)
    t_stat_baseline = t.ppf(single_tailed_confidence_level, degrees_of_freedom_baseline)
    t_stat_reporting = t.ppf(
        single_tailed_confidence_level, degrees_of_freedom_reporting
    )

    rmse_base_residuals_baseline = float(totals_metrics_baseline.rmse_adj)
    rmse_base_residuals_reporting = float(totals_metrics_reporting.rmse_adj)
    autocorr_resid_baseline = totals_metrics_baseline.autocorr_resid
    autocorr_resid_reporting = totals_metrics_reporting.autocorr_resid

    base_avg_baseline = float(totals_metrics_baseline.observed_mean)
    base_avg_reporting = float(totals_metrics_reporting.observed_mean)

    # these result in division by zero error for fsu_error_band
    if (
        post_obs_baseline == 0
        or autocorr_resid_baseline is None
        or abs(autocorr_resid_baseline) == 1
        or base_obs_baseline == 0
        or base_avg_baseline == 0
        or post_obs_reporting == 0
        or autocorr_resid_reporting is None
        or abs(autocorr_resid_reporting) == 1
        or base_obs_reporting == 0
        or base_avg_reporting == 0
    ):
        return None

    autocorr_resid_baseline = float(autocorr_resid_baseline)
    autocorr_resid_reporting = float(autocorr_resid_reporting)

    nprime_baseline = float(
        base_obs_baseline
        * (1 - autocorr_resid_baseline)
        / (1 + autocorr_resid_baseline)
    )
    nprime_reporting = float(
        base_obs_reporting
        * (1 - autocorr_resid_reporting)
        / (1 + autocorr_resid_reporting)
    )

    total_base_energy_baseline = float(base_avg_baseline * base_obs_baseline)
    total_base_energy_reporting = float(base_avg_reporting * base_obs_reporting)

    fsu_error_band_baseline = _compute_fsu_error(
        t_stat_baseline,
        interval_baseline,
        post_obs_baseline,
        total_base_energy_baseline,
        rmse_base_residuals_baseline,
        base_avg_baseline,
        base_obs_baseline,
        nprime_baseline,
    )

    fsu_error_band_reporting = _compute_fsu_error(
        t_stat_reporting,
        interval_reporting,
        post_obs_reporting,
        total_base_energy_reporting,
        rmse_base_residuals_reporting,
        base_avg_reporting,
        base_obs_reporting,
        nprime_reporting,
    )

    return {
        "FSU Error Band: Baseline": fsu_error_band_baseline,
        "FSU Error Band: Reporting": fsu_error_band_reporting,
        "FSU Error Band": (fsu_error_band_baseline**2.0 + fsu_error_band_reporting**2.0)
        ** 0.5,
    }


def modeled_savings(
    baseline_model,
    reporting_model,
    result_index,
    temperature_data,
    with_disaggregated=False,
    confidence_level=0.90,
    predict_kwargs=None,
    degc: bool = False,
):
    """Compute modeled savings, i.e., savings in which baseline and reporting
    usage values are based on models. This is appropriate for annualizing or
    weather normalizing models.

    Parameters
    ----------
    baseline_model : :any:`eemeter.CalTRACKUsagePerDayCandidateModel`
        Model to use for predicting pre-intervention usage.
    reporting_model : :any:`eemeter.CalTRACKUsagePerDayCandidateModel`
        Model to use for predicting post-intervention usage.
    result_index : :any:`pandas.DatetimeIndex`
        The dates for which usage should be modeled.
    temperature_data : :any:`pandas.Series`
        Hourly-frequency timeseries of temperature data during the modeled
        period.
    with_disaggregated : :any:`bool`, optional
        If True, calculate modeled disaggregated usage estimates and savings.
    confidence_level : :any:`float`, optional
        The two-tailed confidence level used to calculate the t-statistic used
        in calculation of the error bands.

        Ignored if not computing error bands.
    predict_kwargs : :any:`dict`, optional
        Extra kwargs to pass to the baseline_model.predict and
        reporting_model.predict methods.
    degc : :any 'bool'
        Relevant temperature units; defaults to False (i.e. Fahrenheit).


    Returns
    -------
    results : :any:`pandas.DataFrame`
        DataFrame with modeled savings, indexed with the result_index. Will
        include the following columns:

        - ``modeled_baseline_usage``
        - ``modeled_reporting_usage``
        - ``modeled_savings``

        If `with_disaggregated` is set to True, the following columns will also
        be in the results DataFrame:

        - ``modeled_baseline_base_load``
        - ``modeled_baseline_cooling_load``
        - ``modeled_baseline_heating_load``
        - ``modeled_reporting_base_load``
        - ``modeled_reporting_cooling_load``
        - ``modeled_reporting_heating_load``
        - ``modeled_base_load_savings``
        - ``modeled_cooling_load_savings``
        - ``modeled_heating_load_savings``
    error_bands : :any:`dict`, optional
        If baseline_model and reporting_model are instances of
        CalTRACKUsagePerDayModelResults, will also return a dictionary of
        FSU and error bands for the aggregated energy savings over the
        normal year period.
    """
    if degc == True:
        temperature_data = 32 + (temperature_data * 1.8)

    prediction_index = result_index

    if predict_kwargs is None:
        predict_kwargs = {}

    model_type = None  # generic
    if isinstance(baseline_model, DailyModel) or isinstance(
        reporting_model, DailyModel
    ):
        raise NotImplementedError(
            "Use predict() with daily and billing models to compute modeled savings."
        )

    def _predicted_usage(model):
        model_prediction = model.predict(
            prediction_index, temperature_data, **predict_kwargs
        )
        predicted_usage = model_prediction.result
        return predicted_usage

    predicted_baseline_usage = _predicted_usage(baseline_model)
    predicted_reporting_usage = _predicted_usage(reporting_model)
    modeled_baseline_usage = predicted_baseline_usage["predicted_usage"].to_frame(
        "modeled_baseline_usage"
    )
    modeled_reporting_usage = predicted_reporting_usage["predicted_usage"].to_frame(
        "modeled_reporting_usage"
    )

    def modeled_savings_func(row):
        return row.modeled_baseline_usage - row.modeled_reporting_usage

    results = modeled_baseline_usage.join(modeled_reporting_usage).assign(
        modeled_savings=modeled_savings_func
    )

    results = results.dropna().reindex(results.index)  # carry NaNs

    error_bands = None
    return results, error_bands



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numpy as np


def neg_log_likelihood(loss, N):
    """
    This function calculates the negative log likelihood for least squares fitting.

    Parameters:
    loss (float): The sum of squared residuals.
    N (int): The number of data points.

    Returns:
    float: The negative log likelihood of the least squares fit.
    """

    # log likelihood for n independent identical normal distributions:
    # log_likelihood = -n/2*np.log(2*np.pi) - n/2*np.log(sigma**2) - 1/(2*sigma**2)*np.sum((x - mu)**2)

    if loss <= 0 or N <= 0:
        return np.inf

    # log likelihood for for least squares fitting:

    res = -N / 2 * (np.log(2 * np.pi) + np.log(loss / N) + 1)

    return res


def selection_criteria(
    loss,
    TSS,
    N,
    num_coeffs,
    model_selection_criteria="bic",
    penalty_multiplier=1.0,
    penalty_power=1.0,
):
    """
    This function calculates the selection criteria for a given model. There are different criteria that can be used,
    and the default is the Bayesian information criterion (BIC).

    Parameters:
    loss (float): The loss of the model.
    TSS (float): Total sum of squares.
    N (int): The number of observations.
    num_coeffs (int): The number of coefficients in the model.
    model_selection_criteria (str): The model selection criteria to use. Default is "bic".
    penalty_multiplier (float): The penalty multiplier. Default is 1.0.
    penalty_power (float): The penalty power. Default is 1.0.

    Returns:
    float: The calculated selection criteria.

    Raises:
    NotImplementedError: If the model selection criteria is "dic", "waic", or "wbic", as these are not implemented.
    """

    K = num_coeffs  # total number of coefficients
    c0 = penalty_multiplier
    d0 = penalty_power

    df_penalized = N - K - 1
    if df_penalized <= 0:
        df_penalized = 1e-6

    # Root-mean-square error adjusted
    if model_selection_criteria.lower() == "rmse":
        criteria = np.sqrt(loss / N)

    # Root-mean-square error adjusted
    elif model_selection_criteria.lower() == "rmse_adj":
        criteria = np.sqrt(loss / df_penalized)

    # 1 - R_squared (because we minimize)
    elif model_selection_criteria.lower() == "r_squared":
        r_squared = 1 - loss / TSS

        criteria = (1 - r_squared) * 100

    elif model_selection_criteria.lower() == "r_squared_adj":
        r_squared = 1 - loss / TSS
        r_squared_adj = 1 - (1 - r_squared) * ((N - 1) / df_penalized)
        criteria = (1 - r_squared_adj) * 100

    # Final prediction error
    elif model_selection_criteria.lower() == "fpe":
        criteria = loss * (N + K + 1) / df_penalized
        # penalized_loss = np.exp(-2/N*log_likelihood)*(N + K)/(N - K)

    # Akaike (ah-kah-ee-kay)
    # Akaike information criterion - Akaike (1973, 1974, 1981)
    elif model_selection_criteria.lower() == "aic":
        criteria = -2 * neg_log_likelihood(loss, N) + c0 * 2 * K**d0

    # Akaike information criterion corrected - Hurvich and Tsai (1989)
    elif model_selection_criteria.lower() == "aicc":
        criteria = (
            -2 * neg_log_likelihood(loss, N)
            + c0 * (2 * K + (2 * K * (K + 1) / df_penalized)) ** d0
        )

    # Consistent Akaike information criterion - Bozdogan (1987)
    elif model_selection_criteria.lower() == "caic":
        criteria = -2 * neg_log_likelihood(loss, N) + c0 * K * (np.log(N) + 1) ** d0

    # Bayesian information criterion
    elif model_selection_criteria.lower() == "bic":
        # if c0 = 0.299 and d0 = 2.1, this is the same as Liu, We, Zidek
        criteria = -2 * neg_log_likelihood(loss, N) + c0 * K * np.log(N) ** d0

    # Sample-size adjusted Bayesian information criteria
    elif model_selection_criteria.lower() == "sabic":
        criteria = (
            -2 * neg_log_likelihood(loss, N) + c0 * K * np.log((N + 2) / 24) ** d0
        )

    # Deviance information criterion
    elif model_selection_criteria.lower() == "dic":
        raise NotImplementedError(
            "DIC has not been implmented as a model selection criterion"
        )

    # Widely applicable (or Watanabe-Akaike) information criterion
    elif model_selection_criteria.lower() == "waic":
        raise NotImplementedError(
            "WAIC has not been implmented as a model selection criterion"
        )

    # Widely applicable (or Watanabe) Bayesian information criterion
    elif model_selection_criteria.lower() == "wbic":
        raise NotImplementedError(
            "WBIC has not been implmented as a model selection criterion"
        )

    if model_selection_criteria.lower() not in ["rmse", "rmse_adj"]:
        criteria /= N  # Normalize to number of datapoints

    return criteria



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from opendsm.eemeter.models.billing.data import (
    BillingBaselineData,
    BillingReportingData,
)
from opendsm.eemeter.samples import load_sample
import numpy as np
import pandas as pd
from pandas import Timestamp, DatetimeIndex, DataFrame
import pytest

TEMPERATURE_SEED = 29
METER_SEED = 41
NUM_DAYS_IN_YEAR = 365


@pytest.fixture
def get_datetime_index(request):
    # Request = [frequency , is_timezone_aware]

    # Create a DateTimeIndex at given frequency and timezone if requested
    inclusive = "both" if request.param[0] in ["MS", "2MS"] else "left"
    datetime_index = pd.date_range(
        start="2023-01-01",
        end="2024-01-01",
        inclusive=inclusive,
        freq=request.param[0],
        tz="US/Eastern" if request.param[1] else None,
    )

    return datetime_index


@pytest.fixture
def get_datetime_index_half_hourly_with_timezone():
    # Create a DateTimeIndex at 30-minute intervals
    datetime_index = pd.date_range(
        start="2023-01-01",
        end="2024-01-01",
        inclusive="left",
        freq="30min",
        tz="US/Eastern",
    )

    return datetime_index


@pytest.fixture
def get_datetime_index_hourly_with_timezone():
    # Create a DateTimeIndex at 30-minute intervals
    datetime_index = pd.date_range(
        start="2023-01-01",
        end="2024-01-01",
        inclusive="left",
        freq="h",
        tz="US/Eastern",
    )

    return datetime_index


@pytest.fixture
def get_datetime_index_daily_with_timezone():
    # Create a DateTimeIndex at daily intervals
    datetime_index = pd.date_range(
        start="2023-01-01",
        end="2024-01-01",
        inclusive="left",
        freq="D",
        tz="US/Eastern",
    )

    return datetime_index


@pytest.fixture
def get_datetime_index_monthly_with_timezone():
    # Create a DateTimeIndex at daily intervals
    datetime_index = pd.date_range(
        start="2023-01-01",
        end="2024-01-01",
        inclusive="both",
        freq="MS",
        tz="US/Eastern",
    )

    return datetime_index


@pytest.fixture
def get_datetime_index_bimonthly_with_timezone():
    # Create a DateTimeIndex at daily intervals
    datetime_index = pd.date_range(
        start="2023-01-01",
        end="2024-01-01",
        inclusive="both",
        freq="2MS",
        tz="US/Eastern",
    )

    return datetime_index


@pytest.fixture
def get_datetime_index_daily_without_timezone():
    # Create a DateTimeIndex at daily intervals
    datetime_index = pd.date_range(
        start="2023-01-01", end="2024-01-01", inclusive="left", freq="D"
    )

    return datetime_index


@pytest.fixture
def get_temperature_data_half_hourly(get_datetime_index_half_hourly_with_timezone):
    datetime_index = get_datetime_index_half_hourly_with_timezone

    np.random.seed(TEMPERATURE_SEED)
    # Create a 'temperature_mean' column with random data
    temperature_mean = np.random.rand(len(datetime_index))

    # Create the DataFrame
    df = pd.DataFrame(data={"temperature": temperature_mean}, index=datetime_index)

    return df


@pytest.fixture
def get_temperature_data_hourly(get_datetime_index_hourly_with_timezone):
    datetime_index = get_datetime_index_hourly_with_timezone

    np.random.seed(TEMPERATURE_SEED)
    # Create a 'temperature_mean' column with random data
    temperature_mean = np.random.rand(len(datetime_index))

    # Create the DataFrame
    df = pd.DataFrame(data={"temperature": temperature_mean}, index=datetime_index)

    return df


@pytest.fixture
def get_temperature_data_daily(get_datetime_index_daily_with_timezone):
    datetime_index = get_datetime_index_daily_with_timezone

    np.random.seed(TEMPERATURE_SEED)
    # Create a 'temperature_mean' column with random data
    temperature_mean = np.random.rand(len(datetime_index))

    # Create the DataFrame
    df = pd.DataFrame(data={"temperature": temperature_mean}, index=datetime_index)

    return df


@pytest.fixture
def get_meter_data_daily(get_datetime_index_daily_with_timezone):
    datetime_index = get_datetime_index_daily_with_timezone

    np.random.seed(METER_SEED)
    # Create a 'meter_value' column with random data
    meter_value = np.random.rand(len(datetime_index))

    # Create the DataFrame
    df = pd.DataFrame(data={"observed": meter_value}, index=datetime_index)

    return df


@pytest.fixture
def get_meter_data_monthly(get_datetime_index_monthly_with_timezone):
    datetime_index = get_datetime_index_monthly_with_timezone

    np.random.seed(METER_SEED)
    # Create a 'meter_value' column with random data
    meter_value = np.random.rand(len(datetime_index))

    # Create the DataFrame
    df = pd.DataFrame(data={"observed": meter_value}, index=datetime_index)
    df.iloc[-1, df.columns.get_loc("observed")] = np.nan

    return df


@pytest.fixture
def get_meter_data_bimonthly(get_datetime_index_bimonthly_with_timezone):
    datetime_index = get_datetime_index_bimonthly_with_timezone

    np.random.seed(METER_SEED)
    # Create a 'meter_value' column with random data
    meter_value = np.random.rand(len(datetime_index))

    # Create the DataFrame
    df = pd.DataFrame(data={"observed": meter_value}, index=datetime_index)
    df.iloc[-1, df.columns.get_loc("observed")] = np.nan

    return df


# Check that a missing timezone raises a Value Error
@pytest.mark.parametrize("get_datetime_index", [["D", False]], indirect=True)
def test_billing_baseline_data_with_missing_timezone(get_datetime_index):
    datetime_index = get_datetime_index

    np.random.seed(TEMPERATURE_SEED)
    # Create a 'temperature_mean' and meter_value columns with random data
    temperature_mean = np.random.rand(len(datetime_index))

    np.random.seed(METER_SEED)
    meter_value = np.random.rand(len(datetime_index))

    # Create the DataFrame
    df = pd.DataFrame(
        data={"meter": meter_value, "temperature": temperature_mean},
        index=datetime_index,
    )

    with pytest.raises(ValueError):
        cls = BillingBaselineData(df, is_electricity_data=True)


# Check that a missing datetime index and column raises a Value Error
def test_billing_baseline_data_with_missing_datetime_index_and_column():
    np.random.seed(TEMPERATURE_SEED)
    # Create a 'temperature_mean' and meter_value columns with random data
    temperature_mean = np.random.rand(NUM_DAYS_IN_YEAR)

    np.random.seed(METER_SEED)
    meter_value = np.random.rand(NUM_DAYS_IN_YEAR)

    # Create the DataFrame
    df = pd.DataFrame(data={"meter": meter_value, "temperature": temperature_mean})

    with pytest.raises(ValueError):
        cls = BillingBaselineData(df, is_electricity_data=True)


@pytest.mark.parametrize("get_datetime_index", [["MS", True]], indirect=True)
def test_billing_baseline_data_with_monthly_frequencies(get_datetime_index):
    datetime_index = get_datetime_index

    np.random.seed(TEMPERATURE_SEED)
    # Create a 'temperature_mean' and meter_value columns with random data
    temperature_mean = np.random.rand(len(datetime_index))

    np.random.seed(METER_SEED)
    meter_value = np.random.rand(len(datetime_index))
    meter_value[-1] = np.nan

    # Create the DataFrame
    df = pd.DataFrame(
        data={"observed": meter_value, "temperature": temperature_mean},
        index=datetime_index,
    )
    df.index = df.index[:-1].union([df.index[-1] - pd.Timedelta(days=1)])

    cls = BillingBaselineData(df, is_electricity_data=True)

    assert cls.df is not None
    assert len(cls.df) == NUM_DAYS_IN_YEAR
    assert round(cls.df.observed.sum(), 2) == round(df.observed.sum(), 2)
    assert len(cls.warnings) == 1
    assert (
        cls.warnings[0].qualified_name
        == "eemeter.sufficiency_criteria.unable_to_confirm_daily_temperature_sufficiency"
    )
    # DQ because only 12 days worth of temperature data is available
    assert len(cls.disqualification) == 2
    assert [dq.qualified_name for dq in cls.disqualification] == [
        "eemeter.sufficiency_criteria.too_many_days_with_missing_data",
        "eemeter.sufficiency_criteria.too_many_days_with_missing_temperature_data",
    ]


@pytest.mark.parametrize("get_datetime_index", [["2MS", True]], indirect=True)
def test_billing_baseline_data_with_bimonthly_frequencies(get_datetime_index):
    datetime_index = get_datetime_index

    np.random.seed(TEMPERATURE_SEED)
    # Create a 'temperature_mean' and meter_value columns with random data
    temperature_mean = np.random.rand(len(datetime_index))

    np.random.seed(METER_SEED)
    meter_value = np.random.rand(len(datetime_index))

    # Create the DataFrame
    df = pd.DataFrame(
        data={"observed": meter_value, "temperature": temperature_mean},
        index=datetime_index,
    )
    df.index = df.index[:-1].union([df.index[-1] - pd.Timedelta(days=1)])
    df.iloc[-1, df.columns.get_loc("observed")] = np.nan

    cls = BillingBaselineData(df, is_electricity_data=True)

    assert cls.df is not None
    # Because two months are missing
    assert len(cls.df) == NUM_DAYS_IN_YEAR
    assert round(cls.df.observed.sum(), 2) == round(df.observed.sum(), 2)
    assert len(cls.warnings) == 1
    assert (
        cls.warnings[0].qualified_name
        == "eemeter.sufficiency_criteria.unable_to_confirm_daily_temperature_sufficiency"
    )
    # DQ because only 6 days worth of temperature data is available
    assert len(cls.disqualification) == 2
    assert set([dq.qualified_name for dq in cls.disqualification]) == set(
        [
            "eemeter.sufficiency_criteria.too_many_days_with_missing_data",
            "eemeter.sufficiency_criteria.too_many_days_with_missing_temperature_data",
        ]
    )


def test_billing_baseline_data_with_monthly_hourly_frequencies(
    get_meter_data_monthly, get_temperature_data_hourly
):
    # Create a DataFrame with uneven frequency
    df = get_temperature_data_hourly

    # Create a DataFrame with daily frequency
    df_meter = get_meter_data_monthly

    # Merge 'df' and 'df_meter' in an outer join
    df = df.merge(df_meter, left_index=True, right_index=True, how="outer")
    df = df[:-1]  # when using dataframe input, rows are exact length

    cls = BillingBaselineData(df, is_electricity_data=True)

    assert cls.df is not None
    assert len(cls.df) == NUM_DAYS_IN_YEAR
    assert round(cls.df.observed.sum(), 2) == round(df.observed.sum(), 2)
    assert len(cls.warnings) == 0
    assert len(cls.disqualification) == 0


def test_billing_baseline_data_with_bimonthly_hourly_frequencies(
    get_meter_data_bimonthly, get_temperature_data_hourly
):
    # Create a DataFrame with uneven frequency
    df = get_temperature_data_hourly

    # Create a DataFrame with daily frequency
    df_meter = get_meter_data_bimonthly

    # Merge 'df' and 'df_meter' in a left join, as df input should not have trailing nan
    df = df.merge(df_meter, left_index=True, right_index=True, how="left")

    cls = BillingBaselineData(df, is_electricity_data=True)

    assert cls.df is not None
    assert len(cls.df) == NUM_DAYS_IN_YEAR
    assert round(cls.df.observed.sum(), 2) == round(df.observed.sum(), 2)
    assert len(cls.warnings) == 0
    assert len(cls.disqualification) == 0


def test_billing_baseline_data_with_monthly_daily_frequencies(
    get_meter_data_monthly, get_temperature_data_daily
):
    # Create a DataFrame with uneven frequency
    df = get_temperature_data_daily

    # Create a DataFrame with daily frequency
    df_meter = get_meter_data_monthly

    # Merge 'df' and 'df_meter' in an outer join
    df = df.merge(df_meter, left_index=True, right_index=True, how="outer")
    df = df[:-1]  # when using dataframe input, rows are exact length

    cls = BillingBaselineData(df, is_electricity_data=True)

    assert cls.df is not None
    assert len(cls.df) == NUM_DAYS_IN_YEAR
    assert round(cls.df.observed.sum(), 2) == round(df.observed.sum(), 2)
    assert len(cls.warnings) == 1
    assert (
        cls.warnings[0].qualified_name
        == "eemeter.sufficiency_criteria.unable_to_confirm_daily_temperature_sufficiency"
    )
    assert len(cls.disqualification) == 0


def test_billing_baseline_data_with_bimonthly_daily_frequencies(
    get_meter_data_bimonthly, get_temperature_data_daily
):
    # Create a DataFrame with uneven frequency
    df = get_temperature_data_daily

    # Create a DataFrame with daily frequency
    df_meter = get_meter_data_bimonthly

    # Merge 'df' and 'df_meter' in an outer join
    df = df.merge(df_meter, left_index=True, right_index=True, how="outer")
    df = df[:-1]  # when using dataframe input, rows are exact length

    cls = BillingBaselineData(df, is_electricity_data=True)

    assert cls.df is not None
    assert len(cls.df) == NUM_DAYS_IN_YEAR
    # assert round(cls.df.observed.sum(), 2) == round(df.observed[:-1].sum(), 2)
    assert round(cls.df.observed.sum(), 2) == round(df.observed.sum(), 2)
    assert len(cls.warnings) == 1
    assert (
        cls.warnings[0].qualified_name
        == "eemeter.sufficiency_criteria.unable_to_confirm_daily_temperature_sufficiency"
    )
    assert len(cls.disqualification) == 0


def test_billing_baseline_data_with_specific_hourly_input():
    meter, temperature, _ = load_sample("il-electricity-cdd-hdd-hourly")
    # Take the extra month for billing data
    meter = meter[
        (meter.index.year == 2017)
        | ((meter.index.year == 2018) & (meter.index.month == 1))
    ]
    temperature = temperature[
        (temperature.index.year == 2017)
        | ((temperature.index.year == 2018) & (temperature.index.month == 1))
    ]

    cls = BillingBaselineData.from_series(meter, temperature, is_electricity_data=True)

    assert cls.df is not None
    assert (
        len(cls.df) == (meter.index[-1] - meter.index[0]).days + 1
    )  # hourly series does not have trailing nan
    assert round(cls.df.observed.sum(), 2) == round(meter.value.sum(), 2)
    assert len(cls.warnings) == 2
    assert [warning.qualified_name for warning in cls.warnings] == [
        "eemeter.data_quality.utc_index",
        "eemeter.sufficiency_criteria.inferior_model_usage",
    ]
    assert len(cls.disqualification) == 1
    assert (
        cls.disqualification[0].qualified_name
        == "eemeter.sufficiency_criteria.incorrect_number_of_total_days"
    )


def test_billing_baseline_data_with_specific_daily_input():
    meter, temperature, _ = load_sample("il-electricity-cdd-hdd-daily")
    # Take the extra month for billing data
    meter = meter[
        (meter.index.year == 2017)
        | ((meter.index.year == 2018) & (meter.index.month == 1))
    ]
    temperature = temperature[
        (temperature.index.year == 2017)
        | ((temperature.index.year == 2018) & (temperature.index.month == 1))
    ]
    cls = BillingBaselineData.from_series(meter, temperature, is_electricity_data=True)

    assert cls.df is not None
    assert (
        len(cls.df) == (meter.index[-1] - meter.index[0]).days + 1
    )  # daily series does not have trailing nan
    assert round(cls.df.observed.sum(), 2) == round(meter.value.sum(), 2)
    assert len(cls.warnings) == 2
    assert [warning.qualified_name for warning in cls.warnings] == [
        "eemeter.data_quality.utc_index",
        "eemeter.sufficiency_criteria.inferior_model_usage",
    ]
    assert len(cls.disqualification) == 1
    assert (
        cls.disqualification[0].qualified_name
        == "eemeter.sufficiency_criteria.incorrect_number_of_total_days"
    )


def test_billing_baseline_data_with_specific_missing_daily_input():
    meter, temperature, _ = load_sample("il-electricity-cdd-hdd-daily")
    # Take the extra month for billing data
    meter = meter[
        (meter.index.year == 2017)
        | ((meter.index.year == 2018) & (meter.index.month == 1))
    ]
    temperature = temperature[
        (temperature.index.year == 2017)
        | ((temperature.index.year == 2018) & (temperature.index.month == 1))
    ]
    # Set 1 month meter data to NaN
    meter.loc[meter.index.month == 4] = np.nan

    cls = BillingBaselineData.from_series(meter, temperature, is_electricity_data=True)

    assert cls.df is not None
    assert (
        len(cls.df) == (meter.index[-1] - meter.index[0]).days + 1
    )  # daily series does not have trailing nan
    assert round(cls.df.observed.sum(), 2) == round(meter.value.sum(), 2)
    assert len(cls.warnings) == 2
    assert [warning.qualified_name for warning in cls.warnings] == [
        "eemeter.data_quality.utc_index",
        "eemeter.sufficiency_criteria.inferior_model_usage",
    ]
    assert len(cls.disqualification) == 1
    assert (
        cls.disqualification[0].qualified_name
        == "eemeter.sufficiency_criteria.incorrect_number_of_total_days"
    )


def test_billing_baseline_data_with_specific_monthly_input():
    meter, temperature, _ = load_sample("il-electricity-cdd-hdd-billing_monthly")
    # Take the extra month for billing data
    meter = meter[
        (meter.index.year == 2017)
        | ((meter.index.year == 2018) & (meter.index.month == 1))
    ]
    temperature = temperature[
        (temperature.index.year == 2017)
        | ((temperature.index.year == 2018) & (temperature.index.month == 1))
    ]
    cls = BillingBaselineData.from_series(meter, temperature, is_electricity_data=True)

    assert cls.df is not None
    assert len(cls.df) == (meter.index[-1] - meter.index[0]).days
    assert round(cls.df.observed.sum(), 2) == round(meter.value.sum(), 2)
    assert len(cls.warnings) == 1
    assert set([warning.qualified_name for warning in cls.warnings]) == set(
        ["eemeter.data_quality.utc_index"]
    )
    assert len(cls.disqualification) == 0


@pytest.mark.parametrize(
    "get_datetime_index", [["30min", True], ["h", True]], indirect=True
)
def test_billing_reporting_data_with_missing_half_hourly_frequencies(
    get_datetime_index,
):
    datetime_index = get_datetime_index
    datetime_index = datetime_index[datetime_index.year == 2023]

    np.random.seed(TEMPERATURE_SEED)
    # Create a 'temperature_mean' and meter_value columns with random data
    temperature_mean = np.random.rand(len(datetime_index))

    # Create the DataFrame
    df = pd.DataFrame(data={"temperature": temperature_mean}, index=datetime_index)

    # Create a mask for Tuesdays and Thursdays
    mask = df.index.dayofweek.isin([1, 3])

    # Set 60% of the temperature data as missing on Tuesdays and Thursdays
    # This should cause the high frequency temperature check to fail on these days
    df.loc[df[mask].sample(frac=0.6, random_state=42).index, "temperature"] = np.nan

    cls = BillingReportingData(df, is_electricity_data=True)

    assert cls.df is not None
    assert len(cls.df) == NUM_DAYS_IN_YEAR

    if datetime_index.freq == "30min":
        assert len(cls.df.temperature.dropna()) == 268
    elif datetime_index.freq == "h":
        assert len(cls.df.temperature.dropna()) == 270

    assert len(cls.warnings) == 1
    assert (
        cls.warnings[0].qualified_name
        == "eemeter.sufficiency_criteria.missing_high_frequency_temperature_data"
    )
    assert len(cls.disqualification) == 3
    expected_disqualifications = [
        "eemeter.sufficiency_criteria.missing_monthly_temperature_data",
        "eemeter.sufficiency_criteria.too_many_days_with_missing_data",
        "eemeter.sufficiency_criteria.too_many_days_with_missing_temperature_data",
    ]
    assert all(
        disqualification.qualified_name in expected_disqualifications
        for disqualification in cls.disqualification
    )


@pytest.mark.parametrize("get_datetime_index", [["D", True]], indirect=True)
def test_billing_reporting_data_with_missing_daily_frequencies(get_datetime_index):
    datetime_index = get_datetime_index
    datetime_index = datetime_index[datetime_index.year == 2023]

    np.random.seed(TEMPERATURE_SEED)
    # Create a 'temperature_mean' and meter_value columns with random data
    temperature_mean = np.random.rand(len(datetime_index))

    # Create the DataFrame
    df = pd.DataFrame(data={"temperature": temperature_mean}, index=datetime_index)

    # Create a mask for Tuesdays and Thursdays
    mask = df.index.dayofweek.isin([1, 3])

    # Set 60% of the temperature data as missing on Tuesdays and Thursdays
    # This should cause the high frequency temperature check to fail on these days
    df.loc[df[mask].sample(frac=0.6, random_state=42).index, "temperature"] = np.nan

    cls = BillingReportingData(df, is_electricity_data=True)

    assert cls.df is not None
    assert len(cls.df) == NUM_DAYS_IN_YEAR
    assert len(cls.df.temperature.dropna()) == len(df.temperature.dropna())
    assert len(cls.warnings) == 1
    assert (
        cls.warnings[0].qualified_name
        == "eemeter.sufficiency_criteria.unable_to_confirm_daily_temperature_sufficiency"
    )
    assert len(cls.disqualification) == 3
    expected_disqualifications = [
        "eemeter.sufficiency_criteria.missing_monthly_temperature_data",
        "eemeter.sufficiency_criteria.too_many_days_with_missing_data",
        "eemeter.sufficiency_criteria.too_many_days_with_missing_temperature_data",
    ]
    assert all(
        disqualification.qualified_name in expected_disqualifications
        for disqualification in cls.disqualification
    )


def test_dst_handling():
    # 2020-03-08 02:00 is nonexistent, should push to 03:00
    tz = "America/New_York"
    idx = DatetimeIndex(
        [
            Timestamp("2020-03-07 02", tz=tz),
            Timestamp("2020-04-06 02", tz=tz),
            Timestamp("2020-05-06 02", tz=tz),
        ]
    )
    df = DataFrame({"observed": [1] * 3, "temperature": [50] * 3}, index=idx)
    baseline = BillingBaselineData(df, is_electricity_data=True)
    assert len(baseline.df) == 61
    hours = np.unique(baseline.df.index.hour)
    assert (hours == [2, 3]).all()

    # 2020-11-01 01:00 is ambiguous, single index should be chosen
    tz = "America/New_York"
    idx = DatetimeIndex(
        [
            Timestamp("2020-10-31 01", tz=tz),
            Timestamp("2020-11-28 01", tz=tz),
            Timestamp("2020-12-28 01", tz=tz),
        ]
    )
    df = DataFrame({"observed": [1] * 3, "temperature": [50] * 3}, index=idx)
    baseline = BillingBaselineData(df, is_electricity_data=True)
    assert (baseline.df.index.hour == 1).all()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from typing import Optional

import numba
import numpy as np

from opendsm.common.adaptive_loss import adaptive_weights
from opendsm.eemeter.models.daily.base_models.full_model import (
    full_model,
    full_model_weight,
)
from opendsm.eemeter.models.daily.objective_function import obj_fcn_decorator
from opendsm.eemeter.models.daily.utilities.opt_settings import OptimizationSettings
from opendsm.eemeter.models.daily.optimize import InitialGuessOptimizer, Optimizer
from opendsm.eemeter.models.daily.parameters import ModelCoefficients, ModelType
from opendsm.eemeter.models.daily.utilities.base_model import (
    fix_identical_bnds,
    get_intercept,
    get_slope,
    get_smooth_coeffs,
)


def fit_hdd_tidd_cdd(
    T,
    obs,
    weights,
    settings,
    opt_options,
    smooth,
    x0: Optional[ModelCoefficients] = None,
    bnds=None,
    initial_fit=False,
):
    # assert x0 is None or x0.model_type is ModelType.HDD_TIDD_CDD_SMOOTH

    if initial_fit:
        alpha = settings.alpha_selection
    else:
        alpha = settings.alpha_final

    if x0 is None:
        x0 = _hdd_tidd_cdd_smooth_x0(T, obs, alpha, settings, smooth)

    max_slope = np.max([x0.hdd_beta, x0.cdd_beta])
    if max_slope != 0:
        max_slope += 10 ** (
            np.log10(np.abs(max_slope)) + np.log10(settings.maximum_slope_oom_scalar)
        )

    if initial_fit:
        T_min = np.min(T)
        T_max = np.max(T)
    else:
        N_min = settings.segment_minimum_count

        T_min = np.partition(T, N_min)[N_min]
        T_max = np.partition(T, -N_min)[-N_min]

    c_hdd_bnds = [T_min, T_max]
    c_hdd_beta_bnds = [0, np.abs(max_slope)]
    intercept_bnds = np.quantile(obs, [0.01, 0.99])
    if smooth:
        c_hdd_k_bnds = [0, 1]
        bnds_0 = [
            c_hdd_bnds,
            c_hdd_beta_bnds,
            c_hdd_k_bnds,
            c_hdd_bnds,
            c_hdd_beta_bnds,
            c_hdd_k_bnds,
            intercept_bnds,
        ]
    else:
        bnds_0 = [
            c_hdd_bnds,
            c_hdd_beta_bnds,
            c_hdd_bnds,
            c_hdd_beta_bnds,
            intercept_bnds,
        ]

    bnds = _hdd_tidd_cdd_smooth_update_bnds(bnds, bnds_0, smooth)

    if smooth:
        coef_id = [
            "hdd_bp",
            "hdd_beta",
            "hdd_k",
            "cdd_bp",
            "cdd_beta",
            "cdd_k",
            "intercept",
        ]
        model_fcn = evaluate_hdd_tidd_cdd_smooth
        weight_fcn = _hdd_tidd_cdd_smooth_weight
        TSS_fcn = None
    else:
        coef_id = ["hdd_bp", "hdd_beta", "cdd_bp", "cdd_beta", "intercept"]
        model_fcn = _hdd_tidd_cdd
        weight_fcn = _hdd_tidd_cdd_weight
        TSS_fcn = _hdd_tidd_cdd_total_sum_of_squares

    obj_fcn = obj_fcn_decorator(
        model_fcn, weight_fcn, TSS_fcn, T, obs, weights, settings, alpha, coef_id, initial_fit
    )

    res = Optimizer(
        obj_fcn, x0.to_np_array(), bnds, coef_id, settings, opt_options
    ).run()

    return res


@numba.jit(nopython=True, error_model="numpy", cache=True)
def _hdd_tidd_cdd(
    hdd_bp,
    hdd_beta,
    cdd_bp,
    cdd_beta,
    intercept,
    T_fit_bnds=np.array([]),
    T=np.array([]),
):
    hdd_k = cdd_k = 0

    return full_model(
        hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept, T_fit_bnds, T
    )


@numba.jit(nopython=True, error_model="numpy", cache=True)
def _hdd_tidd_cdd_smooth(*args):
    return full_model(*args)


def evaluate_hdd_tidd_cdd_smooth(
    hdd_bp,
    hdd_beta,
    hdd_k,
    cdd_bp,
    cdd_beta,
    cdd_k,
    intercept,
    T_fit_bnds,
    T,
    pct_k=True,
):
    if pct_k:
        [hdd_bp, hdd_k, cdd_bp, cdd_k] = get_smooth_coeffs(hdd_bp, hdd_k, cdd_bp, cdd_k)

    return _hdd_tidd_cdd_smooth(
        hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept, T_fit_bnds, T
    )


def _hdd_tidd_cdd_smooth_x0(T, obs, alpha, settings, smooth, min_weight=0.0):
    min_T_idx = settings.segment_minimum_count
    lasso_a = settings.regularization_alpha

    idx_sorted = np.argsort(T).flatten()
    T = T[idx_sorted]
    obs = obs[idx_sorted]

    N = len(obs)

    T_fit_bnds = np.array([T[0], T[-1]])

    def bp_obj_fcn_dec(T, obs, min_T_idx):
        def lasso_penalty(X, wRMSE):
            X_lasso = np.array(X).copy()

            T_range = T_fit_bnds[1] - T_fit_bnds[0]

            X_lasso = np.array(
                [np.min(np.abs(X[idx] - T_fit_bnds)) for idx in range(len(X))]
            )
            X_lasso += (X[1] - X[0]) / 2
            X_lasso *= wRMSE / T_range

            return lasso_a * np.linalg.norm(X_lasso, 1)

        def bp_obj_fcn(x, grad=[], optimize_flag=True):
            if len(x) == 1:
                hdd_bp = cdd_bp = x[0]
            else:
                if x[0] < x[1]:
                    [hdd_bp, cdd_bp] = x
                else:
                    [cdd_bp, hdd_bp] = x

            hdd_beta, cdd_beta, intercept = estimate_betas_and_intercept(
                T, obs, hdd_bp, cdd_bp, min_T_idx, alpha
            )
            hdd_k = cdd_k = 0

            model = _hdd_tidd_cdd_smooth(
                hdd_bp,
                hdd_beta,
                hdd_k,
                cdd_bp,
                cdd_beta,
                cdd_k,
                intercept,
                T_fit_bnds,
                T,
            )
            resid = model - obs

            if alpha == 2:
                resid_mean = np.mean(resid)
                resid -= resid_mean
                intercept += resid_mean
            else:
                resid_median = np.median(resid)
                resid -= resid_median
                intercept += resid_median

            weight, _, _ = adaptive_weights(
                resid, alpha=alpha, sigma=2.698, quantile=0.25, min_weight=min_weight
            )

            loss = np.sum(weight * (resid) ** 2)
            loss += lasso_penalty(x, np.sqrt(loss / N))

            if optimize_flag:
                return loss

            return np.array(
                [hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept]
            )

        return bp_obj_fcn

    obj_fcn = bp_obj_fcn_dec(T, obs, min_T_idx)

    T_bnds = [T[min_T_idx - 1], T[-min_T_idx]]
    if T_bnds[0] == T_bnds[1]:
        T_bnds = [
            np.min(T),
            np.max(T),
        ]  # should be able to do [0] and [-1] but getting error where min > max

    if T_bnds[1] < T_bnds[0]:
        T_bnds = [T_bnds[1], T_bnds[0]]

    T_min = T_bnds[0]
    T_max = T_bnds[1]
    T_range = T_max - T_min

    x0 = np.array([T_range * 0.10, T_range * 0.90]) + T_min
    bnds = np.array([T_bnds, T_bnds])

    opt_settings = OptimizationSettings(
        algorithm=settings.initial_guess_algorithm_choice,
        stop_criteria_type="iteration maximum",
        stop_criteria_value=200,
        initial_step=settings.initial_step_percentage,
        x_tol_rel=1e-3,
        f_tol_rel=0.5,
    )

    res = InitialGuessOptimizer(
        obj_fcn, x0, bnds, opt_settings
    ).run()

    x0 = obj_fcn(res.x, optimize_flag=False)

    if smooth:
        model_type = ModelType.HDD_TIDD_CDD_SMOOTH
        hdd_k = x0[2]
        cdd_k = x0[5]
    else:
        model_type = ModelType.HDD_TIDD_CDD
        hdd_k = cdd_k = None

    return ModelCoefficients(
        model_type=model_type,
        hdd_bp=x0[0],
        hdd_beta=x0[1],
        hdd_k=hdd_k,
        cdd_bp=x0[3],
        cdd_beta=x0[4],
        cdd_k=cdd_k,
        intercept=x0[6],
    )


def estimate_betas_and_intercept(T, obs, hdd_bp, cdd_bp, min_T_idx, alpha):
    idx_hdd = np.argwhere(T < hdd_bp).flatten()
    idx_tidd = np.argwhere((hdd_bp <= T) & (T <= cdd_bp)).flatten()
    idx_cdd = np.argwhere(cdd_bp < T).flatten()

    if len(idx_tidd) > 0:
        intercept = get_intercept(obs[idx_tidd], alpha)
    elif (
        (len(idx_cdd) >= min_T_idx)
        and (len(idx_hdd) >= min_T_idx)
        and (idx_cdd[min_T_idx - 1] - idx_hdd[-min_T_idx]) > 0
    ):
        intercept = get_intercept(
            obs[idx_hdd[-min_T_idx] : idx_cdd[min_T_idx - 1]], alpha
        )
    else:
        intercept = np.quantile(obs, 0.20)

    hdd_beta = get_slope(T[idx_hdd], obs[idx_hdd], hdd_bp, intercept, alpha)
    if hdd_beta > 0:
        hdd_beta = 0
    else:
        hdd_beta *= -1

    cdd_beta = get_slope(T[idx_cdd], obs[idx_cdd], cdd_bp, intercept, alpha)
    if cdd_beta < 0:
        cdd_beta = 0

    return hdd_beta, cdd_beta, intercept


def _hdd_tidd_cdd_smooth_update_bnds(new_bnds, bnds, smooth):
    if new_bnds is None:
        new_bnds = bnds

    # breakpoint bounds
    new_bnds[0] = bnds[0]
    if smooth:
        new_bnds[3] = bnds[3]
    else:
        new_bnds[2] = bnds[2]

    # intercept bounds at index 6 for smooth, 4 for unsmooth
    if smooth:
        new_bnds[6] = bnds[6]
    else:
        new_bnds[4] = bnds[4]

    new_bnds = np.sort(new_bnds, axis=1)
    new_bnds = fix_identical_bnds(new_bnds)

    # beta and k must be non-negative
    if smooth:
        beta_k_idx = [1, 2, 4, 5]
    else:
        beta_k_idx = [1, 3]
    for i in beta_k_idx:
        if new_bnds[i][0] < 0:
            new_bnds[i][0] = 0

    return new_bnds


def _hdd_tidd_cdd_weight(
    hdd_bp,
    hdd_beta,
    cdd_bp,
    cdd_beta,
    intercept,
    T,
    residual,
    sigma=3.0,
    quantile=0.25,
    alpha=2.0,
    min_weight=0.0,
):
    hdd_k = cdd_k = 0
    model_vars = [hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept]

    return full_model_weight(
        *model_vars, T, residual, sigma, quantile, alpha, min_weight
    )


def _hdd_tidd_cdd_smooth_weight(
    hdd_bp,
    hdd_beta,
    hdd_k,
    cdd_bp,
    cdd_beta,
    cdd_k,
    intercept,
    T,
    residual,
    sigma=3.0,
    quantile=0.25,
    alpha=2.0,
    min_weight=0.0,
):
    model_vars = [hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept]

    return full_model_weight(
        *model_vars, T, residual, sigma, quantile, alpha, min_weight
    )


def _hdd_tidd_cdd_total_sum_of_squares(
    hdd_bp, hdd_beta, cdd_bp, cdd_beta, intercept, T, obs
):
    if hdd_bp > cdd_bp:
        hdd_bp, cdd_bp = cdd_bp, hdd_bp

    idx_hdd_bp = np.argmin(np.abs(T - hdd_bp))
    idx_cdd_bp = np.argmin(np.abs(T - cdd_bp))

    TSS = []
    for observed in [obs[:idx_hdd_bp], obs[idx_hdd_bp:idx_cdd_bp], obs[idx_cdd_bp:]]:
        if len(observed) == 0:
            continue

        TSS.append(np.sum((observed - np.mean(observed)) ** 2))

    TSS = np.sum(TSS)

    return TSS



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
__all__ = (
    "EEMeterError",
    "NoBaselineDataError",
    "NoReportingDataError",
    "MissingModelParameterError",
    "UnrecognizedModelTypeError",
    "DataSufficiencyError",
    "DisqualifiedModelError",
)


class EEMeterError(Exception):
    """Base class for EEmeter library errors."""

    pass


class NoBaselineDataError(EEMeterError):
    """Error indicating lack of baseline data."""

    pass


class NoReportingDataError(EEMeterError):
    """Error indicating lack of reporting data."""

    pass


class MissingModelParameterError(EEMeterError):
    """Error indicating missing model parameter."""

    pass


class UnrecognizedModelTypeError(EEMeterError):
    """Error indicating unrecognized model type."""

    pass


class DataSufficiencyError(EEMeterError):
    """Error indicating insufficient data to fit model on."""

    pass


class DisqualifiedModelError(EEMeterError):
    """Error indicating attempt to predict with disqualified or poorly fit model."""

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import json
import numpy as np
import pandas as pd
import pytest

from opendsm.eemeter.models.hourly_caltrack.model import (
    caltrack_hourly_fit_feature_processor,
    caltrack_hourly_prediction_feature_processor,
    fit_caltrack_hourly_model_segment,
    fit_caltrack_hourly_model,
)
from opendsm.eemeter.common.features import (
    compute_time_features,
)


@pytest.fixture
def segmented_data():
    index = pd.date_range(start="2017-01-01", periods=24, freq="h", tz="UTC")
    time_features = compute_time_features(index)
    segmented_data = pd.DataFrame(
        {
            "hour_of_week": time_features.hour_of_week,
            "temperature_mean": np.linspace(0, 100, 24),
            "meter_value": np.linspace(10, 70, 24),
            "weight": np.ones((24,)),
        },
        index=index,
    )
    return segmented_data


@pytest.fixture
def occupancy_lookup():
    index = pd.Categorical(range(168))
    occupancy = pd.Series([i % 2 == 0 for i in range(168)], index=index)
    return pd.DataFrame(
        {"dec-jan-feb-weighted": occupancy, "jan-feb-mar-weighted": occupancy}
    )


@pytest.fixture
def occupied_temperature_bins():
    bins = pd.Series([True, True, True], index=[30, 60, 90])
    return pd.DataFrame({"dec-jan-feb-weighted": bins, "jan-feb-mar-weighted": bins})


@pytest.fixture
def unoccupied_temperature_bins():
    bins = pd.Series([False, True, True], index=[30, 60, 90])
    return pd.DataFrame({"dec-jan-feb-weighted": bins, "jan-feb-mar-weighted": bins})


def test_caltrack_hourly_fit_feature_processor(
    segmented_data,
    occupancy_lookup,
    occupied_temperature_bins,
    unoccupied_temperature_bins,
):
    result = caltrack_hourly_fit_feature_processor(
        "dec-jan-feb-weighted",
        segmented_data,
        occupancy_lookup,
        occupied_temperature_bins,
        unoccupied_temperature_bins,
    )
    assert list(result.columns) == [
        "meter_value",
        "hour_of_week",
        "bin_0_occupied",
        "bin_1_occupied",
        "bin_2_occupied",
        "bin_3_occupied",
        "bin_0_unoccupied",
        "bin_1_unoccupied",
        "bin_2_unoccupied",
        "weight",
    ]
    assert result.shape == (24, 10)
    result.hour_of_week = result.hour_of_week.astype(float)
    assert round(result.sum().sum(), 2) == 5916.0


def test_caltrack_hourly_prediction_feature_processor(
    segmented_data,
    occupancy_lookup,
    occupied_temperature_bins,
    unoccupied_temperature_bins,
):
    result = caltrack_hourly_prediction_feature_processor(
        "dec-jan-feb-weighted",
        segmented_data,
        occupancy_lookup,
        occupied_temperature_bins,
        unoccupied_temperature_bins,
    )
    assert list(result.columns) == [
        "hour_of_week",
        "bin_0_occupied",
        "bin_1_occupied",
        "bin_2_occupied",
        "bin_3_occupied",
        "bin_0_unoccupied",
        "bin_1_unoccupied",
        "bin_2_unoccupied",
        "weight",
    ]
    assert result.shape == (24, 9)
    result.hour_of_week = result.hour_of_week.astype(float)
    assert round(result.sum().sum(), 2) == 4956.0


@pytest.fixture
def segmented_design_matrices(
    segmented_data,
    occupancy_lookup,
    occupied_temperature_bins,
    unoccupied_temperature_bins,
):
    return {
        "dec-jan-feb-weighted": caltrack_hourly_fit_feature_processor(
            "dec-jan-feb-weighted",
            segmented_data,
            occupancy_lookup,
            occupied_temperature_bins,
            unoccupied_temperature_bins,
        )
    }


def test_fit_caltrack_hourly_model_segment(segmented_design_matrices):
    segment_name = "dec-jan-feb-weighted"
    segment_data = segmented_design_matrices[segment_name]
    segment_model = fit_caltrack_hourly_model_segment(segment_name, segment_data)
    assert segment_model.formula == (
        "meter_value ~ C(hour_of_week) - 1 + bin_0_occupied"
        " + bin_1_occupied + bin_2_occupied + bin_3_occupied"
        " + bin_0_unoccupied + bin_1_unoccupied + bin_2_unoccupied"
    )
    assert segment_model.segment_name == "dec-jan-feb-weighted"
    assert len(segment_model.model_params.keys()) == 31
    assert segment_model.model is not None
    assert segment_model.warnings is not None
    prediction = segment_model.predict(segment_data)
    assert round(prediction.sum(), 2) == 960.0


@pytest.fixture
def temps():
    index = pd.date_range(start="2017-01-01", periods=24, freq="h", tz="UTC")
    temps = pd.Series(np.linspace(0, 100, 24), index=index)
    return temps


@pytest.mark.parametrize("segment_type", ["three_month_weighted"])
def test_fit_caltrack_hourly_model(
    segmented_design_matrices,
    occupancy_lookup,
    occupied_temperature_bins,
    unoccupied_temperature_bins,
    temps,
    segment_type,
):
    segmented_model_results = fit_caltrack_hourly_model(
        segmented_design_matrices,
        occupancy_lookup,
        occupied_temperature_bins,
        unoccupied_temperature_bins,
        segment_type=segment_type,
    )

    assert segmented_model_results.model.segment_models is not None
    assert str(segmented_model_results).startswith("CalTRACKHourlyModelResults")
    prediction = segmented_model_results.predict(temps.index, temps).result


@pytest.mark.parametrize("segment_type", ["single", "three_month_weighted"])
def test_serialize_caltrack_hourly_model(
    segmented_design_matrices,
    occupancy_lookup,
    occupied_temperature_bins,
    unoccupied_temperature_bins,
    temps,
    segment_type,
):
    segmented_model = fit_caltrack_hourly_model(
        segmented_design_matrices,
        occupancy_lookup,
        occupied_temperature_bins,
        unoccupied_temperature_bins,
        segment_type=segment_type,
    )
    assert json.dumps(segmented_model.json())


@pytest.fixture
def segmented_data_nans():
    num_periods = 200
    index = pd.date_range(start="2017-01-01", periods=num_periods, freq="h", tz="UTC")
    time_features = compute_time_features(index)
    segmented_data = pd.DataFrame(
        {
            "hour_of_week": time_features.hour_of_week,
            "temperature_mean": np.linspace(0, 100, num_periods),
            "meter_value": np.linspace(10, 70, num_periods),
            "weight": np.ones((num_periods,)),
        },
        index=index,
    )
    return segmented_data


@pytest.fixture
def occupancy_lookup_nans():
    index = pd.Categorical(range(168))
    occupancy = pd.Series([i % 2 == 0 for i in range(168)], index=index)
    occupancy_nans = pd.Series([np.nan for i in range(168)], index=index)
    return pd.DataFrame(
        {
            "dec-jan-feb-weighted": occupancy,
            "jan-feb-mar-weighted": occupancy,
            "apr-may-jun-weighted": occupancy_nans,
        }
    )


@pytest.fixture
def temperature_bins_nans():
    bins = pd.Series([True, True, True], index=[30, 60, 90])
    bins_nans = pd.Series([False, False, False], index=[30, 60, 90])
    return pd.DataFrame(
        {
            "dec-jan-feb-weighted": bins,
            "jan-feb-mar-weighted": bins,
            "apr-may-jun-weighted": bins_nans,
        }
    )


@pytest.fixture
def segmented_design_matrices_nans(
    segmented_data_nans, occupancy_lookup_nans, temperature_bins_nans
):
    return {
        "dec-jan-feb-weighted": caltrack_hourly_fit_feature_processor(
            "dec-jan-feb-weighted",
            segmented_data_nans,
            occupancy_lookup_nans,
            temperature_bins_nans,
            temperature_bins_nans,
        ),
        "apr-may-jun-weighted": caltrack_hourly_fit_feature_processor(
            "apr-may-jun-weighted",
            segmented_data_nans,
            occupancy_lookup_nans,
            temperature_bins_nans,
            temperature_bins_nans,
        ),
    }


@pytest.mark.parametrize("segment_type", ["three_month_weighted"])
def test_fit_caltrack_hourly_model_nans_less_than_week_predict(
    segmented_design_matrices_nans,
    occupancy_lookup_nans,
    temperature_bins_nans,
    temps_extended,
    temps,
    segment_type,
):
    segmented_model_results = fit_caltrack_hourly_model(
        segmented_design_matrices_nans,
        occupancy_lookup_nans,
        temperature_bins_nans,
        temperature_bins_nans,
        segment_type=segment_type,
    )

    assert segmented_model_results.model.segment_models is not None
    assert segmented_model_results.model.model_lookup["jan"].model is not None
    assert segmented_model_results.model.model_lookup["may"].model is not None
    assert segmented_model_results.model.model_lookup["may"].warnings == []
    prediction = segmented_model_results.predict(temps.index, temps).result
    assert prediction.shape[0] == 24
    assert prediction["predicted_usage"].sum().round() == 955.0


@pytest.fixture
def segmented_data_nans_less_than_week():
    num_periods = 4
    index = pd.date_range(start="2017-01-01", periods=num_periods, freq="h", tz="UTC")
    time_features = compute_time_features(index)
    segmented_data = pd.DataFrame(
        {
            "hour_of_week": time_features.hour_of_week,
            "temperature_mean": np.linspace(0, 100, num_periods),
            "meter_value": np.linspace(10, 70, num_periods),
            "weight": np.ones((num_periods,)),
        },
        index=index,
    )
    return segmented_data


@pytest.fixture
def occupancy_lookup_nans_less_than_week():
    index = pd.Categorical(range(168))
    occupancy = pd.Series([i % 2 == 0 for i in range(168)], index=index)
    occupancy_nans_less_than_week = pd.Series([np.nan for i in range(168)], index=index)
    return pd.DataFrame(
        {
            "dec-jan-feb-weighted": occupancy,
            "jan-feb-mar-weighted": occupancy,
            "apr-may-jun-weighted": occupancy_nans_less_than_week,
        }
    )


@pytest.fixture
def temperature_bins_nans_less_than_week():
    bins = pd.Series([True, True, True], index=[30, 60, 90])
    bins_nans_less_than_week = pd.Series([False, False, False], index=[30, 60, 90])
    return pd.DataFrame(
        {
            "dec-jan-feb-weighted": bins,
            "jan-feb-mar-weighted": bins,
            "apr-may-jun-weighted": bins_nans_less_than_week,
        }
    )


@pytest.fixture
def segmented_design_matrices_nans_less_than_week(
    segmented_data_nans_less_than_week,
    occupancy_lookup_nans_less_than_week,
    temperature_bins_nans_less_than_week,
):
    return {
        "dec-jan-feb-weighted": caltrack_hourly_fit_feature_processor(
            "dec-jan-feb-weighted",
            segmented_data_nans_less_than_week,
            occupancy_lookup_nans_less_than_week,
            temperature_bins_nans_less_than_week,
            temperature_bins_nans_less_than_week,
        ),
        "apr-may-jun-weighted": caltrack_hourly_fit_feature_processor(
            "apr-may-jun-weighted",
            segmented_data_nans_less_than_week,
            occupancy_lookup_nans_less_than_week,
            temperature_bins_nans_less_than_week,
            temperature_bins_nans_less_than_week,
        ),
    }


@pytest.fixture
def temps_extended():
    index = pd.date_range(start="2017-01-01", periods=168, freq="h", tz="UTC")
    temps = pd.Series(1, index=index)
    return temps


@pytest.mark.parametrize("segment_type", ["three_month_weighted"])
def test_fit_caltrack_hourly_model_nans_less_than_week_fit(
    segmented_design_matrices_nans_less_than_week,
    occupancy_lookup_nans_less_than_week,
    temperature_bins_nans_less_than_week,
    temps_extended,
    segment_type,
):
    segmented_model_results = fit_caltrack_hourly_model(
        segmented_design_matrices_nans_less_than_week,
        occupancy_lookup_nans_less_than_week,
        temperature_bins_nans_less_than_week,
        temperature_bins_nans_less_than_week,
        segment_type=segment_type,
    )

    assert segmented_model_results.model.segment_models is not None
    prediction = segmented_model_results.predict(
        temps_extended.index, temps_extended
    ).result
    assert prediction.shape[0] == 168
    assert prediction.dropna().shape[0] == 4


@pytest.fixture
def segmented_design_matrices_empty_models(
    segmented_data,
    occupancy_lookup,
    occupied_temperature_bins,
    unoccupied_temperature_bins,
):
    return {
        "dec-jan-feb-weighted": caltrack_hourly_fit_feature_processor(
            "dec-jan-feb-weighted",
            segmented_data[:0],
            occupancy_lookup,
            occupied_temperature_bins,
            unoccupied_temperature_bins,
        )
    }


@pytest.mark.parametrize("segment_type", ["three_month_weighted"])
def test_predict_caltrack_hourly_model_empty_models(
    temps,
    segmented_design_matrices_empty_models,
    occupancy_lookup,
    occupied_temperature_bins,
    unoccupied_temperature_bins,
    segment_type,
):
    segmented_model_results = fit_caltrack_hourly_model(
        segmented_design_matrices_empty_models,
        occupancy_lookup,
        occupied_temperature_bins,
        unoccupied_temperature_bins,
        segment_type=segment_type,
    )

    assert segmented_model_results.model.segment_models is not None
    prediction = segmented_model_results.predict(temps.index, temps).result
    assert prediction.shape[0] == 24
    assert prediction.dropna().shape[0] == 0


@pytest.fixture
def occupancy_lookup_zeroes():
    index = pd.Categorical(range(168))
    occupancy = pd.Series([False] * 168, index=index)
    return pd.DataFrame(
        {"dec-jan-feb-weighted": occupancy, "jan-feb-mar-weighted": occupancy}
    )


@pytest.fixture
def segmented_design_matrices_single_mode(
    segmented_data,
    occupancy_lookup_zeroes,
    occupied_temperature_bins,
    unoccupied_temperature_bins,
):
    return {
        "dec-jan-feb-weighted": caltrack_hourly_fit_feature_processor(
            "dec-jan-feb-weighted",
            segmented_data,
            occupancy_lookup_zeroes,
            occupied_temperature_bins,
            unoccupied_temperature_bins,
        )
    }


def test_fit_caltrack_hourly_model_segment_single_mode(
    segmented_design_matrices_single_mode,
):
    segment_name = "dec-jan-feb-weighted"
    segment_data = segmented_design_matrices_single_mode[segment_name]
    segment_model = fit_caltrack_hourly_model_segment(segment_name, segment_data)
    assert segment_model.formula == (
        "meter_value ~ C(hour_of_week) - 1 + bin_0_occupied + bin_1_occupied"
        " + bin_2_occupied + bin_3_occupied + bin_0_unoccupied + bin_1_unoccupied"
        " + bin_2_unoccupied"
    )
    assert segment_model.segment_name == "dec-jan-feb-weighted"
    assert len(segment_model.model_params.keys()) == 31
    assert segment_model.model is not None
    assert segment_model.warnings is not None
    prediction = segment_model.predict(segment_data)
    assert round(prediction.sum(), 2) == 960.0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numpy as np
import pandas as pd
import pytest
from opendsm.eemeter.models.daily.parameters import ModelCoefficients

from opendsm.eemeter.models.daily.utilities.settings import DailySettings as Settings
from opendsm.eemeter.models.daily.parameters import ModelType
from opendsm.eemeter.models.daily.fit_base_models import (
    fit_initial_models_from_full_model,
    fit_model,
    fit_final_model,
    _get_opt_settings,
)

from opendsm.eemeter.models.daily.optimize_results import OptimizedResult


@pytest.fixture
def meter_data():
    df_meter = pd.DataFrame(
        {
            "temperature": [
                10.0,
                15.0,
                20.0,
                25.0,
                30.0,
                35.0,
                40.0,
                45.0,
                50.0,
                55.0,
                60.0,
                65.0,
                70.0,
                75.0,
                80.0,
                85.0,
                90.0,
                95.0,
                100.0,
            ],
            "observed": [
                100.0,
                150.0,
                200.0,
                250.0,
                300.0,
                350.0,
                400.0,
                450.0,
                500.0,
                550.0,
                600.0,
                650.0,
                700.0,
                750.0,
                800.0,
                850.0,
                900.0,
                950.0,
                1000.0,
            ],
            "datetime": [
                "2022-01-01",
                "2022-01-02",
                "2022-01-03",
                "2022-01-04",
                "2022-01-05",
                "2022-01-06",
                "2022-01-07",
                "2022-01-08",
                "2022-01-09",
                "2022-01-10",
                "2022-01-11",
                "2022-01-12",
                "2022-01-13",
                "2022-01-14",
                "2022-01-15",
                "2022-01-16",
                "2022-01-17",
                "2022-01-18",
                "2022-01-19",
            ],
        }
    )
    df_meter["datetime"] = pd.to_datetime(df_meter["datetime"])
    df_meter.set_index("datetime", inplace=True)
    return df_meter


@pytest.fixture
def get_settings():
    return Settings()


@pytest.fixture
def get_optimized_result(get_settings):
    return OptimizedResult(
        x=np.array([1, 2, 3, 4]),
        bnds=[
            [0, 1],
            [0, 2],
            [0, 3],
            [0, 4],
            [0, 5],
            [0, 6],
            [0, 7],
            [0, 8],
            [0, 9],
            [0, 10],
        ],
        coef_id=["c_hdd_bp", "c_hdd_beta", "c_hdd_k", "intercept"],
        loss_alpha=0.1,
        C=0.5,
        T=np.array([10, 20, 30, 40, 50, 60, 70, 80, 90, 100]),
        model=np.array([2, 4, 6, 8, 10, 12, 14, 16, 18, 20]),
        weight=np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]),
        resid=np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]),
        jac=None,
        mean_loss=0.2,
        TSS=0.3,
        success=True,
        message="Optimization successful.",
        nfev=10,
        time_elapsed=0.5,
        settings=get_settings,
    )


def test_fit_initial_models_from_full_model(meter_data, get_settings):
    # Test case 1: Test the function with a sample dataset
    model_res = fit_initial_models_from_full_model(meter_data, get_settings)
    assert isinstance(model_res, OptimizedResult)

    # Test case 3: Test the function with a dataset that has missing values
    T = np.array([10, 20, 30, 40, 50])
    obs = np.array([1, 2, np.nan, 4, 5])
    model_res = fit_initial_models_from_full_model(meter_data, get_settings)
    assert isinstance(model_res, OptimizedResult)

    # Test case 4: Test the function with a dataset that has negative values
    T = np.array([10, 20, 30, 40, 50])
    obs = np.array([-1, 2, 3, 4, 5])
    model_res = fit_initial_models_from_full_model(meter_data, get_settings)
    assert isinstance(model_res, OptimizedResult)


def test_fit_model(meter_data, get_settings):
    # Test case 1: Test for model_key = "hdd_tidd_cdd_smooth"
    T = np.array(meter_data["temperature"])
    obs = np.array(meter_data["observed"])
    weights = None

    fit_input = [T, obs, weights, get_settings, _get_opt_settings(get_settings)]
    res = fit_model("hdd_tidd_cdd_smooth", fit_input, None, None)
    assert isinstance(res, OptimizedResult)

    # Test case 2: Test for model_key = "hdd_tidd_cdd"
    res = fit_model("hdd_tidd_cdd", fit_input, None, None)
    assert isinstance(res, OptimizedResult)

    # Test case 3: Test for model_key = "c_hdd_tidd_smooth"
    res = fit_model("c_hdd_tidd_smooth", fit_input, None, None)
    assert isinstance(res, OptimizedResult)

    # Test case 4: Test for model_key = "c_hdd_tidd"
    x0 = ModelCoefficients(
        model_type=ModelType.HDD_TIDD,
        cdd_beta=1.747475624458497,
        cdd_bp=74.69216148926878,
        cdd_k=0.2548934690459498,
        hdd_beta=1.308196391571347,
        hdd_bp=63.332029669746596,
        hdd_k=0.0,
        intercept=49.97929032502437,
    )
    res = fit_model("c_hdd_tidd", fit_input, x0, None)
    assert isinstance(res, OptimizedResult)

    # Test case 5: Test for model_key = "tidd"
    res = fit_model("tidd", fit_input, None, None)
    assert isinstance(res, OptimizedResult)

    # TODO: add more data specific cases


def test_fit_final_model(meter_data, get_settings, get_optimized_result):
    # Test case 1: Test if the function returns an instance of OptimizedResult
    HoF = get_optimized_result

    res = fit_final_model(meter_data, HoF, get_settings)
    assert isinstance(res, OptimizedResult)

    # Test case 2: Test if the function raises a TypeError when the input arguments are of the wrong type
    with pytest.raises(TypeError):
        fit_final_model("not a dataframe", "not an OptimizedResult", "not a dictionary")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import pytest

from opendsm.eemeter.models.daily.utilities.settings import (
    DailySettings,
)


def test_default_settings():
    settings = DailySettings()
    assert settings.developer_mode is False
    assert settings.algorithm_choice.lower() == "nlopt_sbplx"
    assert settings.initial_guess_algorithm_choice.lower() == "nlopt_direct"
    assert settings.alpha_selection == 2.0
    assert settings.alpha_final == "adaptive"
    assert settings.alpha_final_type == "last"
    assert settings.regularization_alpha == 0.001
    assert settings.regularization_percent_lasso == 1.0
    assert settings.allow_smooth_model is True
    assert settings.split_selection.allow_separate_summer is True
    assert settings.split_selection.allow_separate_shoulder is True
    assert settings.split_selection.allow_separate_winter is True
    assert settings.split_selection.allow_separate_weekday_weekend is True
    assert settings.split_selection.reduce_splits_by_gaussian is True
    assert settings.segment_minimum_count == 6


def test_custom_settings():
    settings_dict = {
        "developer_mode": True,
        "algorithm_choice": "scipy_SLSQP",
        "initial_guess_algorithm_choice": "nlopt_DIRECT_L",
        "alpha_selection": 1.5,
        "alpha_final": 1.5,
        "alpha_final_type": "last",
        "regularization_alpha": 0.01,
        "regularization_percent_lasso": 0.5,
        "allow_smooth_model": True,
        "split_selection": {
            "allow_separate_summer": True,
            "allow_separate_shoulder": True,
            "allow_separate_winter": True,
            "allow_separate_weekday_weekend": True,
            "reduce_splits_by_gaussian": True,
        },
        "segment_minimum_count": 20,
    }
    settings = DailySettings(**settings_dict)

    assert settings.developer_mode is True
    assert settings.algorithm_choice.lower() == "scipy_slsqp"
    assert settings.initial_guess_algorithm_choice.lower() == "nlopt_direct_l"
    assert settings.alpha_selection == 1.5
    assert settings.alpha_final == 1.5
    assert settings.alpha_final_type == "last"
    assert settings.regularization_alpha == 0.01
    assert settings.regularization_percent_lasso == 0.5
    assert settings.allow_smooth_model is True
    assert settings.split_selection.allow_separate_summer is True
    assert settings.split_selection.allow_separate_shoulder is True
    assert settings.split_selection.allow_separate_winter is True
    assert settings.split_selection.allow_separate_weekday_weekend is True
    assert settings.split_selection.reduce_splits_by_gaussian is True
    assert settings.segment_minimum_count == 20


def test_invalid_settings():
    with pytest.raises(ValueError):
        DailySettings(developer_mode=False, algorithm_choice="invalid_algorithm")
    with pytest.raises(ValueError):
        DailySettings(developer_mode=False, alpha_selection=0.5)
    with pytest.raises(ValueError):
        DailySettings(developer_mode=False, alpha_selection=1.5)
    with pytest.raises(ValueError):
        DailySettings(developer_mode=False, alpha_final_type="invalid_type")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import pandas as pd

from opendsm.eemeter.common.features import (
    compute_temperature_features,
    compute_time_features,
    compute_usage_per_day_feature,
    merge_features,
)
from opendsm.eemeter.models.hourly_caltrack.model import (
    caltrack_hourly_fit_feature_processor,
)
from opendsm.eemeter.models.hourly_caltrack.segmentation import (
    iterate_segmented_dataset,
)

__all__ = (
    "create_caltrack_hourly_preliminary_design_matrix",
    "create_caltrack_hourly_segmented_design_matrices",
    "create_caltrack_daily_design_matrix",
    "create_caltrack_billing_design_matrix",
)


def create_caltrack_hourly_preliminary_design_matrix(
    meter_data, temperature_data, degc: bool = False
):
    """A helper function which calls basic feature creation methods to create an
    input suitable for use in the first step of creating a CalTRACK hourly model.

    Parameters
    ----------
    meter_data : :any:`pandas.DataFrame`
        Hourly meter data in eemeter format.
    temperature_data : :any:`pandas.Series`
        Hourly temperature data in eemeter format.
    degc : :any 'bool'
        Relevant temperature units; defaults to False (i.e. Fahrenheit).

    Returns
    -------
    design_matrix : :any:`pandas.DataFrame`
        A design matrix with meter_value, hour_of_week, hdd_(hbp_default), and cdd_(cbp_default) features.
    """

    if degc == True:
        temperature_data = 32 + (temperature_data * 1.8)

    time_features = compute_time_features(
        meter_data.index, hour_of_week=True, hour_of_day=False, day_of_week=False
    )
    temperature_features = compute_temperature_features(
        meter_data.index,
        temperature_data,
        heating_balance_points=[50],
        cooling_balance_points=[
            65
        ],  # note both HBP this will require further work in future iterations of eemeter - CBP in particular assumes all buildings have cooling, which is a strong assumption.
        degree_day_method="hourly",
    )
    design_matrix = merge_features(
        [meter_data.value.to_frame("meter_value"), temperature_features, time_features]
    )
    return design_matrix


def create_caltrack_billing_design_matrix(
    meter_data, temperature_data, degc: bool = False
):
    """A helper function which calls basic feature creation methods to create a
    design matrix suitable for use with CalTRACK Billing methods.

    Parameters
    ----------
    meter_data : :any:`pandas.DataFrame`
        Monthly meter data in eemeter format.
    temperature_data : :any:`pandas.Series`
        Hourly temperature data in eemeter format.
    degc : :any 'bool'
        Relevant temperature units; defaults to Fahrenheit.

    Returns
    -------
    design_matrix : :any:`pandas.DataFrame`
        A design matrix with mean usage_per_day and temperature features.
    """
    usage_per_day = compute_usage_per_day_feature(meter_data, series_name="meter_value")
    usage_per_day = usage_per_day.resample("D").ffill()
    if degc == True:
        temperature_data = 32 + (temperature_data * 1.8)

    temperature_features = compute_temperature_features(
        usage_per_day.index,
        temperature_data,
        data_quality=True,
        tolerance=pd.Timedelta(
            "35D"
        ),  # limit temperature data matching to periods of up to 35 days.
    )
    design_matrix = merge_features([usage_per_day, temperature_features])
    return design_matrix


def create_caltrack_daily_design_matrix(
    meter_data, temperature_data, degc: bool = False
):
    """A helper function which calls basic feature creation methods to create a
    design matrix suitable for use with CalTRACK daily methods.

    Parameters
    ----------
    meter_data : :any:`pandas.DataFrame`
        Daily meter data in eemeter format.
    temperature_data : :any:`pandas.Series`
        Hourly temperature data in eemeter format.
     degc : :any 'bool'
        Relevant temperature units; defaults to Fahrenheit.

    Returns
    -------
    design_matrix : :any:`pandas.DataFrame`
        A design matrix with mean usage_per_day and temperature features
    """
    usage_per_day = compute_usage_per_day_feature(meter_data, series_name="meter_value")
    if degc == True:
        temperature_data = 32 + (temperature_data * 1.8)

    temperature_features = compute_temperature_features(
        meter_data.index,
        temperature_data,
        data_quality=True,
    )
    design_matrix = merge_features([usage_per_day, temperature_features])
    return design_matrix


def create_caltrack_hourly_segmented_design_matrices(
    preliminary_design_matrix,
    segmentation,
    occupancy_lookup,
    occupied_temperature_bins,
    unoccupied_temperature_bins,
):
    """A helper function which calls basic feature creation methods to create a
    design matrix suitable for use with segmented CalTRACK hourly models.
    Parameters
    ----------
    preliminary_design_matrix : :any:`pandas.DataFrame`
        A dataframe of the form returned by
        :any:`eemeter.create_caltrack_hourly_preliminary_design_matrix`.
    segmentation : :any:`pandas.DataFrame`
        Weights for each segment. This is a dataframe of the form returned by
        :any:`eemeter.segment_time_series` on the `preliminary_design_matrix`.
    occupancy_lookup : any:`pandas.DataFrame`
        Occupancy for each segment. This is a dataframe of the form returned by
        :any:`eemeter.estimate_hour_of_week_occupancy`.
    occupied_temperature_bins : :any:``
        Occupied temperature bin settings for each segment. This is a dataframe of the
        form returned by :any:`eemeter.fit_temperature_bins`.
    unoccupied_temperature_bins : :any:``
        Ditto, for unoccupied.
    Returns
    -------
    design_matrix : :any:`dict` of :any:`pandas.DataFrame`
        A dict of design matrixes created using the
        :any:`eemeter.caltrack_hourly_fit_feature_processor`.
    """
    return {
        segment_name: segmented_data
        for segment_name, segmented_data in iterate_segmented_dataset(
            preliminary_design_matrix,
            segmentation=segmentation,
            feature_processor=caltrack_hourly_fit_feature_processor,
            feature_processor_kwargs={
                "occupancy_lookup": occupancy_lookup,
                "occupied_temperature_bins": occupied_temperature_bins,
                "unoccupied_temperature_bins": unoccupied_temperature_bins,
            },
        )
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from opendsm.eemeter.common.exceptions import (
    EEMeterError,
    NoBaselineDataError,
    NoReportingDataError,
    MissingModelParameterError,
    UnrecognizedModelTypeError,
)

import pytest


def test_eemeter_error():
    with pytest.raises(EEMeterError):
        raise EEMeterError


def test_no_baseline_data_error():
    with pytest.raises(NoBaselineDataError):
        raise NoBaselineDataError
    assert isinstance(NoBaselineDataError(), EEMeterError)


def test_no_reporting_data_error():
    with pytest.raises(NoReportingDataError):
        raise NoReportingDataError
    assert isinstance(NoReportingDataError(), EEMeterError)


def test_missing_model_parameter_error():
    with pytest.raises(MissingModelParameterError):
        raise MissingModelParameterError
    assert isinstance(MissingModelParameterError(), EEMeterError)


def test_unrecognized_model_type_error():
    with pytest.raises(UnrecognizedModelTypeError):
        raise UnrecognizedModelTypeError
    assert isinstance(UnrecognizedModelTypeError(), EEMeterError)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from abc import ABC, abstractmethod


class AbstractDataProcessor(ABC):
    """Abstract class for data processors."""

    @abstractmethod
    def set_data(self, data):
        """Process data.

        Parameters
        ----------
        data : pd.DataFrame
            Data to process.

        Returns
        -------
        processed_data : pd.DataFrame
            Processed data.
        """
        pass

    @abstractmethod
    def _check_data_sufficiency(self, data):
        """Check data sufficiency.

        Parameters
        ----------
        data : pd.DataFrame
            Data to check.

        Returns
        -------
        is_sufficient : bool
            Whether the data is sufficient.
        """
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import json
import importlib.resources

import pytest

from opendsm.eemeter.samples import load_sample


@pytest.fixture
def sample_metadata():
    with importlib.resources.files("opendsm.eemeter.samples").joinpath(
        "metadata.json"
    ).open("rb") as f:
        metadata = json.loads(f.read().decode("utf-8"))
    return metadata


def _from_sample(sample, tempF=True):
    meter_data, temperature_data, metadata = load_sample(sample, tempF=tempF)
    return {
        "meter_data": meter_data,
        "temperature_data": temperature_data,
        "blackout_start_date": metadata["blackout_start_date"],
        "blackout_end_date": metadata["blackout_end_date"],
    }


@pytest.fixture
def il_electricity_cdd_hdd_hourly():
    return _from_sample("il-electricity-cdd-hdd-hourly")


@pytest.fixture
def il_electricity_cdd_hdd_daily():
    return _from_sample("il-electricity-cdd-hdd-daily")


@pytest.fixture
def il_electricity_cdd_hdd_billing_monthly():
    return _from_sample("il-electricity-cdd-hdd-billing_monthly")


@pytest.fixture
def il_electricity_cdd_hdd_billing_bimonthly():
    return _from_sample("il-electricity-cdd-hdd-billing_bimonthly")


@pytest.fixture
def il_gas_hdd_only_hourly():
    return _from_sample("il-gas-hdd-only-hourly")


@pytest.fixture
def uk_electricity_hdd_only_hourly_sample_1():
    return _from_sample("uk-electricity-hdd-only-hourly-sample-1", tempF=False)


@pytest.fixture
def uk_electricity_hdd_only_hourly_sample_2():
    return _from_sample("uk-electricity-hdd-only-hourly-sample-2", tempF=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numba
import numpy as np
from scipy.optimize import minimize_scalar

from opendsm.common.adaptive_loss_Z import ln_Z
from opendsm.common.utils import OoM_numba

LOSS_ALPHA_MIN = -100.0


@numba.jit(nopython=True, cache=True)
def weighted_quantile(
    values, quantiles, weights=None, values_presorted=False, old_style=False
):
    """https://stackoverflow.com/questions/21844024/weighted-percentile-using-numpy
    Very close to numpy.percentile, but supports weights.
    NOTE: quantiles should be in [0, 1]!
    :param values: numpy.array with data
    :param quantiles: array-like with many quantiles needed
    :param sample_weight: array-like of the same length as `array`
    :param values_presorted: bool, if True, then will avoid sorting of
        initial array
    :param old_style: if True, will correct output to be consistent
        with numpy.quantile.
    :return: numpy.array with computed quantiles.
    """
    finite_idx = np.where(np.isfinite(values))
    values = values[finite_idx]
    if weights is None or len(weights) == 0:
        weights = np.ones_like(values)
    else:
        weights = weights[finite_idx]

    assert np.all(quantiles >= 0) and np.all(
        quantiles <= 1
    ), "quantiles should be in [0, 1]"

    if not values_presorted:
        sorter = np.argsort(values)
        values = values[sorter]
        weights = weights[sorter]

    res = np.cumsum(weights) - 0.5 * weights
    if old_style:  # To be convenient with numpy.quantile
        res -= res[0]
        res /= res[-1]
    else:
        res /= np.sum(weights)

    return np.interp(quantiles, res, values)


def remove_outliers(data, weights=None, sigma_threshold=3, quantile=0.25):
    outlier_bnds = IQR_outlier(data, weights, sigma_threshold, quantile)
    idx_no_outliers = np.argwhere(
        (data >= outlier_bnds[0]) & (data <= outlier_bnds[1])
    ).flatten()
    data_no_outliers = data[idx_no_outliers]

    return data_no_outliers, idx_no_outliers


@numba.jit(nopython=True, cache=True)
def IQR_outlier(data, weights=None, sigma_threshold=3, quantile=0.25):
    # only use finite data
    if weights is None:
        q13 = np.nanquantile(data[np.isfinite(data)], [quantile, 1 - quantile])
    else:  # weighted_quantile could be used always, don't know speed
        q13 = weighted_quantile(
            data[np.isfinite(data)], np.array([quantile, 1 - quantile]), weights=weights
        )

    q13_scalar = (
        0.7413 * sigma_threshold - 0.5
    )  # this is a pretty good fit to get the scalar for any sigma
    iqr = np.diff(q13)[0] * q13_scalar
    outlier_threshold = np.array([q13[0] - iqr, q13[1] + iqr])

    return outlier_threshold


@numba.jit(nopython=True, cache=True)
def sliding_window(
    arr, window_size, step=0
):  # https://giov.dev/2018/05/a-window-on-numpy-s-views.html
    """Assuming a time series with time advancing along dimension 0,
        window the time series with given size and step.

    :param arr : input array.
    :type arr: numpy.ndarray
    :param window_size: size of sliding window.
    :type window_size: int
    :param step: step size of sliding window. If 0, step size is set to obtain
        non-overlapping contiguous windows (that is, step=window_size).
        Defaults to 0.
    :type step: int

    :return: array
    :rtype: numpy.ndarray
    """
    n_obs = arr.shape[0]

    # validate arguments
    if window_size > n_obs:
        raise ValueError(
            "Window size must be less than or equal "
            "the size of array in first dimension."
        )
    if step < 0:
        raise ValueError("Step must be positive.")

    n_windows = 1 + int(np.floor((n_obs - window_size) / step))

    obs_stride = arr.strides[0]
    windowed_row_stride = obs_stride * step

    new_shape = (n_windows, window_size) + arr.shape[1:]
    new_strides = (windowed_row_stride,) + arr.strides

    strided = np.lib.stride_tricks.as_strided(
        arr,
        shape=new_shape,
        strides=new_strides,
    )
    return strided


def rolling_IQR_outlier(x, y, sigma_threshold=3, quantile=0.25, window=0.05, step=1):
    """
    This function calculates the outlier threshold using the rolling Interquartile Range (IQR) method.

    Parameters:
    x (numpy array): The x-values of the data.
    y (numpy array): The y-values of the data.
    sigma_threshold (float, optional): The sigma threshold for the IQR. Default is 3.
    quantile (float, optional): The quantile to use for the IQR calculation. Default is 0.25.
    window (float or int, optional): The window size for the rolling calculation. If less than 1, it is considered as a proportion of the data length. Default is 0.05.
    step (int, optional): The step size for the rolling calculation. Default is 1.

    Returns:
    numpy array: A 2D array where the first row is the lower outlier threshold and the second row is the upper outlier threshold.
    """

    if window <= 1.0:
        window = int(np.floor(len(y) * window))
    else:
        window = int(window)

    step = int(np.floor(window * step))
    if step < 1:
        step = 1

    y = np.abs(y)

    x_windows = sliding_window(x, window, step=step)
    y_windows = sliding_window(y, window, step=step)

    x_interp = np.zeros(np.shape(x_windows)[0])
    q1 = np.zeros(np.shape(y_windows)[0])
    q3 = np.zeros(np.shape(y_windows)[0])
    for i in range(len(q1)):
        x_interp[i] = np.mean(x_windows[i])

        q13 = np.quantile(y_windows[i], [quantile, 1 - quantile])
        q1[i] = q13[0]
        q3[i] = q13[1]

    q13_scalar = (
        0.7413 * sigma_threshold - 0.5
    )  # this is a pretty good fit to get the scalar for any sigma
    iqr = (q3 - q1) * q13_scalar
    outlier_bnds = [q1 - iqr, q3 + iqr]

    outlier_threshold = np.zeros((2, len(x)))
    outlier_threshold[0] = np.interp(x, x_interp, outlier_bnds[0])
    outlier_threshold[1] = np.interp(x, x_interp, outlier_bnds[1])

    # x_interp = np.arange(0, len(outlier_bnds[0]))
    # x_orig = np.linspace(0, len(outlier_bnds[0]), len(x))

    # outlier_threshold = np.zeros((2, len(x)))
    # outlier_threshold[0] = np.interp(x_orig, x_interp, outlier_bnds[0])
    # outlier_threshold[1] = np.interp(x_orig, x_interp, outlier_bnds[1])

    return outlier_threshold


@numba.jit(nopython=True, error_model="numpy", cache=True)
def get_C(resid, mu, sigma, quantile=0.25, algo="iqr_legacy"):
    """
    This function calculates the maximum absolute value of the Interquartile Range (IQR) of the residuals
    from a given mean (mu) and standard deviation (sigma). If the maximum absolute value is zero,
    it is replaced with the order of magnitude of the maximum value of the IQR.

    Parameters:
    resid (numpy.ndarray): The residuals of the model.
    mu (float): The mean of the residuals.
    sigma (float): The standard deviation of the residuals.
    quantile (float, optional): The quantile to use for the IQR calculation. Default is 0.25.

    Returns:
    float: The maximum absolute value of the IQR of the residuals, or the order of magnitude of the maximum value of the IQR.
    """
    # remove non-finite values
    resid = resid[np.isfinite(resid)]

    if algo == "iqr_legacy":
        # TODO: uncertain if these C functions should use np.min, np.mean, or np.max
        # suspect we can switch to IQR below, but need to test
        bounds = IQR_outlier(
            resid - mu, weights=None, sigma_threshold=sigma, quantile=quantile
        )
        C = np.max(np.abs(bounds))

    elif algo == "iqr":
        resid = np.abs(resid)

        bounds = IQR_outlier(
            resid - mu, weights=None, sigma_threshold=sigma, quantile=quantile
        )
        C = np.max(np.abs(bounds))

    elif algo == "mad":
        resid_median = np.median(resid)
        MAD = np.median(np.abs(resid - resid_median))

        C = sigma * 1.4826 * MAD

    elif algo == "stdev":
        C = sigma * np.std(resid)

    if C == 0:
        C = OoM_numba(np.array([C]), method="floor")[0]

    return C


def rolling_C(T, resid, mu, sigma=3, quantile=0.25, window=0.2, step=1.0):
    """
    Function to calculate the rolling C value.

    Parameters:
    T (array-like): The time series data.
    resid (array-like): The residuals of the model.
    mu (float): The mean value.
    sigma (float, optional): The standard deviation. Default is 3.
    quantile (float, optional): The quantile to use for the IQR calculation. Default is 0.25.
    window (float, optional): The window size for the rolling calculation. Default is 0.2.
    step (float, optional): The step size for the rolling calculation. Default is 1.0.

    Returns:
    C (array-like): The calculated rolling C values.
    """

    q13 = rolling_IQR_outlier(T, resid - mu, sigma, quantile, window, step)
    C = np.max(np.abs(q13), axis=0)

    return C


@numba.jit(nopython=True, error_model="numpy", cache=True)
def generalized_loss_fcn(x, a=2, a_min=LOSS_ALPHA_MIN):
    """
    This function calculates the generalized loss function based on the given parameters.

    Parameters:
    x (float): The input value for which the loss is to be calculated.
    a (float): The parameter that determines the type of loss function to be used. Default is 2.
    a_min (float): The minimum value of 'a' for which the Welsch/Leclerc loss is to be calculated. Default is LOSS_ALPHA_MIN.

    Returns:
    float: The calculated loss.

    The function supports the following types of loss functions based on the value of 'a', i.e. alpha:
    - L2 loss for a=2.0
    - Smoothed L1 loss for a=1.0
    - Charbonnier loss for a=0.0
    - Cauchy/Lorentzian loss for a=-2.0
    - Welsch/Leclerc loss for a<=a_min
    - Generalized loss for other values of 'a'
    """

    # defaults to sum of squared error
    x_2 = x**2

    if a == 2.0:  # L2
        loss = 0.5 * x_2
    elif a == 1.0:  # smoothed L1
        loss = np.sqrt(x_2 + 1) - 1
    elif a == 0.0:  # Charbonnier loss
        loss = np.log(0.5 * x_2 + 1)
    elif a == -2.0:  # Cauchy/Lorentzian loss
        loss = 2 * x_2 / (x_2 + 4)
    elif a <= a_min:  # at -infinity, Welsch/Leclerc loss
        loss = 1 - np.exp(-0.5 * x_2)
    else:
        loss = np.abs(a - 2) / a * ((x_2 / np.abs(a - 2) + 1) ** (a / 2) - 1)

    return loss


@numba.jit(nopython=True, error_model="numpy", cache=True)
def generalized_loss_derivative(x, c=1, a=2):
    """
    This function calculates the derivative of the generalized loss function.

    Parameters:
    x (float): The input value for which the derivative of the loss function is to be calculated.
    c (float, optional): The scale parameter. Default is 1.
    a (float, optional): Alpha value for the loss function. Default is 2.

    Returns:
    float: The derivative of the generalized loss function at x.

    The function supports different types of loss functions based on the value of alpha:
    - L2 loss for a=2.0
    - Smoothed L1 loss for a=1.0
    - Charbonnier loss for a=0.0
    - Cauchy/Lorentzian loss for a=-2.0
    - Welsch/Leclerc loss for a<=LOSS_ALPHA_MIN
    - Generalized loss for other values of alpha
    """

    if a == 2.0:  # L2
        dloss_dx = x / c**2
    elif a == 1.0:  # smoothed L1
        dloss_dx = x / c**2 / np.sqrt((x / c) ** 2 + 1)
    elif a == 0.0:  # Charbonnier loss
        dloss_dx = 2 * x / (x**2 + 2 * c**2)
    elif a == -2.0:  # Cauchy/Lorentzian loss
        dloss_dx = 16 * c**2 * x / (4 * c**2 + x**2) ** 2
    elif a <= LOSS_ALPHA_MIN:  # at -infinity, Welsch/Leclerc loss
        dloss_dx = x / c**2 * np.exp(-0.5 * (x / c) ** 2)
    else:
        dloss_dx = x / c**2 * ((x / c) ** 2 / np.abs(a - 2) + 1)

    return dloss_dx


@numba.jit(nopython=True, error_model="numpy", cache=True)
def generalized_loss_weights(x: np.ndarray, a: float = 2, min_weight: float = 0.00):
    """
    This function calculates the generalized loss weights for a given array of values.

    Parameters:
    x (np.ndarray): The input array for which the weights are to be calculated.
    a (float, optional): The alpha parameter for the weight calculation. Default is 2.
    min_weight (float, optional): The minimum weight to be considered. Default is 0.00.

    Returns:
    np.ndarray: The calculated weights for the input array.
    """

    dtype = numba.float64
    if numba.config.DISABLE_JIT:
        dtype = np.float64

    w = np.ones(len(x), dtype=dtype)
    for i, xi in enumerate(x):
        if a == 2 or xi <= 0:
            w[i] = 1
        elif a == 0:
            w[i] = 1 / (0.5 * xi**2 + 1)
        elif a <= LOSS_ALPHA_MIN:
            w[i] = np.exp(-0.5 * xi**2)
        else:
            w[i] = (xi**2 / np.abs(a - 2) + 1) ** (0.5 * a - 1)

    return w * (1 - min_weight) + min_weight


# penalize the loss function using approximate partition function
# default to L2 loss
def penalized_loss_fcn(x, a=2, use_penalty=True):
    """
    This function calculates the penalized loss for a given input. There is a penalty for more coefficients
    being added into the funciton being fit.

    Parameters:
    x (numpy array): The input data for which the loss is to be calculated.
    a (float, optional): The parameter for the generalized loss function. Default is 2.
    use_penalty (bool, optional): Flag to decide whether to use penalty or not. Default is True.

    Returns:
    loss (float): The calculated penalized loss.

    Raises:
    Exception: If non-finite values are found in the calculated loss.
    """

    loss = generalized_loss_fcn(x, a)

    if use_penalty:
        penalty = ln_Z(
            a, LOSS_ALPHA_MIN
        )  # approximate partition function for C=1, tau=10
        loss += penalty

        if not np.isfinite(loss).all():
            # print("a: ", a)
            # print("x: ", x)
            # print("penalty: ", penalty)
            raise Exception("non-finite values in 'penalized_loss_fcn'")

    return loss


@numba.jit(nopython=True, error_model="numpy", cache=True)
def alpha_scaled(s, a_max=2):
    """
    This function calculates the alpha value based on the input s and a_max (max alpha allowed).

    Parameters:
    s (float): The input value for which alpha is to be calculated.
    a_max (int, optional): The maximum value of alpha. Default is 2.

    Returns:
    float: The calculated alpha value.

    Note:
    The function uses two different methods to calculate alpha based on the value of a_max.
    If a_max is 2, it uses a specific formula involving logarithmic and exponential functions.
    If a_max is not 2, it uses a different formula also involving logarithmic and exponential functions.
    """

    if a_max == 2:
        a = 3
        b = 0.25

        if s < 0:
            s = 0

        if s > 1:
            s = 1

        s_max = 1 - 2 / (1 + 10**a)
        s = (1 - 2 / (1 + 10 ** (a * s**b))) / s_max

        alpha = LOSS_ALPHA_MIN + (2 - LOSS_ALPHA_MIN) * s

    else:
        x0 = 1
        k = 1.5  # 1 or 1.5, testing required

        if s >= 1:
            return 100
        elif s <= 0:
            return -100

        A = (np.exp((100 - x0) / k) + 1) / (1 - np.exp(200 / k))
        K = (1 - A) * np.exp((x0 - 100) / k) + 1

        alpha = x0 - k * np.log((K - A) / (s - A) - 1)

    return alpha


def adaptive_loss_fcn(x, mu=0, c=1, alpha="adaptive", replace_nonfinite=True):
    """
    This function calculates the adaptive loss function value and the alpha parameter.

    Parameters:
    x (numpy array): The input data.
    mu (float, optional): The mean value for standardization. Default is 0.
    c (float, optional): The scale value for standardization. Default is 1.
    alpha (str or float, optional): The alpha parameter for the loss function. If "adaptive", the function will optimize alpha. Default is "adaptive".
    replace_nonfinite (bool, optional): If True, replaces non-finite values in x with the maximum finite value. Default is True.

    Returns:
    tuple: The value of the loss function and the alpha parameter.
    """

    if np.all(mu != 0) or np.all(c != 1):
        x = (x - mu) / c  # standardized residuals

    if replace_nonfinite:
        x[~np.isfinite(x)] = np.max(x[np.isfinite(x)])

    loss_alpha_fcn = lambda alpha: penalized_loss_fcn(
        x, a=alpha, use_penalty=True
    ).sum()

    if alpha == "adaptive":  #
        res = minimize_scalar(
            lambda s: loss_alpha_fcn(alpha_scaled(s)),
            bounds=[-1e-5, 1 + 1e-5],
            method="Bounded",
            options={"xatol": 1e-5},
        )
        loss_alpha = alpha_scaled(res.x)
        # res = minimize(lambda s: loss_alpha_fcn(alpha_scaled(s[0])), x0=[0.7], bounds=[[0, 1]], method="L-BFGS-B")
        # loss_alpha = alpha_scaled(res.x[0])
        loss_fcn_val = res.fun

    else:
        loss_alpha = alpha
        loss_fcn_val = loss_alpha_fcn(alpha)

    return loss_fcn_val, loss_alpha


# Assumes that x has not been standardized
def adaptive_weights(
    x, 
    alpha="adaptive", 
    sigma=3, 
    quantile=0.25, 
    min_weight=0.00, 
    C_algo="iqr_legacy", 
    replace_nonfinite=True
):
    """
    This function calculates the adaptive weights for a given data set.

    Parameters:
    x (numpy.array): The input data.
    alpha (str or float, optional): The alpha value for the adaptive loss function. Default is "adaptive".
    sigma (float, optional): The sigma threshold for outlier removal. Default is 3.
    quantile (float, optional): The quantile for outlier removal. Default is 0.25.
    min_weight (float, optional): The minimum weight. Default is 0.00.
    replace_nonfinite (bool, optional): Whether to replace non-finite values. Default is True.

    Returns:
    tuple: A tuple containing the generalized loss weights, C value, and alpha value.
    """

    x_no_outlier, _ = remove_outliers(x, sigma_threshold=sigma, quantile=0.25)

    # TODO: Should x be abs or not?
    # likely should be abs
    # mu = np.median(np.abs(x_no_outlier))
    mu = np.median(x_no_outlier)

    C = get_C(x, mu, sigma, quantile, C_algo)
    x = (x - mu) / C

    if alpha == "adaptive":
        _, alpha = adaptive_loss_fcn(
            x, alpha=alpha, replace_nonfinite=replace_nonfinite
        )

    return generalized_loss_weights(x, a=alpha, min_weight=min_weight), C, alpha



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from __future__ import annotations

import datetime

import numpy as np
import pandas as pd
from pandas.tseries.offsets import MonthBegin, MonthEnd

from opendsm.eemeter.common.data_processor_utilities import (
    as_freq,
    clean_billing_daily_data,
    compute_minimum_granularity,
)
from opendsm.eemeter.common.features import compute_temperature_features
from opendsm.eemeter.common.warnings import EEMeterWarning
from opendsm.eemeter.common.sufficiency_criteria import BillingSufficiencyCriteria
from opendsm.eemeter.models.daily.data import _DailyData

"""TODO there is still a ton of unecessarily duplicated code between billing+daily.
    we should be able to perform a few transforms within the billing baseclass, and then call super() for the rest

    unsure whether we should inherit from the public classes because we'll have to take care to use type(data)
    instead of isinstance(data,  _) when doing the checks in the model/wrapper to avoid unintentionally allowing a mix of data/model type
"""


class _BillingData(_DailyData):
    """Baseline data processor for billing data.

    2.2.3.4. Off-cycle reads (spanning less than 25 days) should be dropped from analysis.
    These readings typically occur due to meter reading problems or changes in occupancy.

    2.2.3.5. For pseudo-monthly billing cycles, periods spanning more than 35 days should be dropped from analysis.
    For bi-monthly billing cycles, periods spanning more than 70 days should be dropped from the analysis.
    """

    def _compute_meter_value_df(self, df: pd.DataFrame):
        """
        Computes the meter value DataFrame by cleaning and processing the observed meter data.
        1. The minimum granularity is computed from the non null rows. If the billing cycle is mixed between monthly and bimonthly, then the minimum granularity is bimonthly
        2. The meter data is cleaned and downsampled/upsampled into the correct frequency using clean_billing_daily_data()
        3. Add missing days as NaN by merging with a full year daily index.

        Parameters
        ----------

            df (pd.DataFrame): The DataFrame containing the observed meter data.

        Returns
        -------
            pd.DataFrame: The cleaned and processed meter value DataFrame.
        """
        meter_series_full = df["observed"]
        meter_series = meter_series_full.dropna()
        if meter_series.empty:
            return meter_series_full.resample("D").first().to_frame()

        start_date = meter_series_full.index.min()
        end_date = meter_series_full.index.max().replace(
            hour=meter_series.index[-1].hour
        )  # assume final period ends on same hour

        # ensure we adjust backwards to normalize hour, never adding time
        if end_date > meter_series_full.index.max():
            end_date = end_date - pd.Timedelta(days=1)

        min_granularity = compute_minimum_granularity(
            meter_series.index, default_granularity="billing_bimonthly"
        )

        # Ensure higher frequency data is aggregated to the monthly model
        if not min_granularity.startswith("billing"):
            # MS is so that the date for Month Start
            meter_series = meter_series.resample("MS").sum(min_count=1)
            # normalize to midnight since we're picking an arbitrary day to represent period start anyway
            end_date = end_date.normalize()
            self.warnings.append(
                EEMeterWarning(
                    qualified_name="eemeter.sufficiency_criteria.inferior_model_usage",
                    description=(
                        "Daily data is provided but the model used is monthly. Are you sure this is the intended model?"
                    ),
                    data={},
                )
            )
            min_granularity = "billing_monthly"

        # Adjust index to follow final nan convention--without this, final period will be short one day
        meter_series[end_date + pd.Timedelta(days=1)] = np.nan

        # This checks for offcycle reads. That is a disqualification if the billing cycle is less than 25 days
        meter_value_df = clean_billing_daily_data(
            meter_series.to_frame("value"), min_granularity, self.disqualification
        )

        # Spread billing data to daily
        meter_value_df = as_freq(meter_value_df["value"], "D").to_frame("value")
        meter_value_df = meter_value_df[:-1]
        meter_value_df = meter_value_df.rename(columns={"value": "observed"})

        # This will ensure that the missing days are kept in the dataframe
        # Create an index with all the days from the start and end date of 'meter_value_df'
        if len(meter_value_df) > 0:
            all_days_index = pd.date_range(
                start=start_date,
                end=end_date,
                freq="D",
                tz=df.index.tz,
                ambiguous=True,
                nonexistent="shift_forward",
            )
            all_days_df = pd.DataFrame(index=all_days_index)
            meter_value_df = meter_value_df.merge(
                all_days_df, left_index=True, right_index=True, how="outer"
            )

        return meter_value_df

    def _compute_temperature_features(
        self, df: pd.DataFrame, meter_index: pd.DatetimeIndex
    ):
        """
        Compute temperature features for the given DataFrame and meter index.
        1. The frequency of the temperature data is inferred and set to hourly if not already. If frequency is not inferred or its lower than hourly, a warning is added.
        2. The temperature data is downsampled/upsampled into the daily frequency using as_freq()
        3. High frequency temperature data is checked for missing values and a warning is added if more than 50% of the data is missing, and those rows are set to NaN.
        4. If frequency was already hourly, compute_temperature_features() is used to recompute the temperature to match with the meter index.

        Parameters
        ----------

            df (pd.DataFrame): The DataFrame containing temperature data.
            meter_index (pd.DatetimeIndex): The meter index.

        Returns
        -------

            pd.Series: The computed temperature values.
            pd.DataFrame: The computed temperature features.
        """
        temp_series = df["temperature"]
        temp_series.index.freq = temp_series.index.inferred_freq
        if temp_series.index.freq != "h":
            if (
                temp_series.index.freq is None
                or isinstance(temp_series.index.freq, MonthEnd)
                or isinstance(temp_series.index.freq, MonthBegin)
                or temp_series.index.freq > pd.Timedelta(hours=1)
            ):
                # Add warning for frequencies longer than 1 hour
                self.warnings.append(
                    EEMeterWarning(
                        qualified_name="eemeter.sufficiency_criteria.unable_to_confirm_daily_temperature_sufficiency",
                        description=(
                            "Cannot confirm that pre-aggregated temperature data had sufficient hours kept"
                        ),
                        data={},
                    )
                )
            # TODO consider disallowing this until a later patch
            if temp_series.index.freq != "D":
                # Downsample / Upsample the temperature data to daily
                temperature_features = as_freq(
                    temp_series, "D", series_type="instantaneous", include_coverage=True
                )
                # If high frequency data check for 50% data coverage in rollup
                if len(temperature_features[temperature_features.coverage <= 0.5]) > 0:
                    self.warnings.append(
                        EEMeterWarning(
                            qualified_name="eemeter.sufficiency_criteria.missing_high_frequency_temperature_data",
                            description=(
                                "More than 50% of the high frequency Temperature data is missing."
                            ),
                            data={
                                "high_frequency_data_missing_count": len(
                                    temperature_features[
                                        temperature_features.coverage <= 0.5
                                    ].index.to_list()
                                )
                            },
                        )
                    )

                # Set missing high frequency data to NaN
                temperature_features.loc[
                    temperature_features.coverage > 0.5, "value"
                ] = (
                    temperature_features[temperature_features.coverage > 0.5].value
                    / temperature_features[temperature_features.coverage > 0.5].coverage
                )

                temperature_features = (
                    temperature_features[temperature_features.coverage > 0.5]
                    .reindex(temperature_features.index)[["value"]]
                    .rename(columns={"value": "temperature_mean"})
                )

                if "coverage" in temperature_features.columns:
                    temperature_features = temperature_features.drop(
                        columns=["coverage"]
                    )
            else:
                temperature_features = temp_series.to_frame(name="temperature_mean")

            temperature_features["temperature_null"] = temp_series.isnull().astype(int)
            temperature_features["temperature_not_null"] = temp_series.notnull().astype(
                int
            )
            temperature_features["n_days_kept"] = 0  # unused
            temperature_features["n_days_dropped"] = 0  # unused
        else:
            if not meter_index.empty:
                buffer_idx = meter_index.max() + pd.Timedelta(days=1)
                meter_index = meter_index.union([buffer_idx])

            temperature_features = compute_temperature_features(
                meter_index,
                temp_series,
                data_quality=True,
            )
            temperature_features = temperature_features[:-1]
            # Only check for high frequency temperature data if it exists
            # TODO this check causes weird behavior with very sparse temp data.
            # will still get DQ'd, but final df receives non-nan temperatures
            median_samples = (
                temperature_features.temperature_not_null
                + temperature_features.temperature_null
            ).median()
            if median_samples > 1:
                invalid_temperature_rows = (
                    temperature_features.temperature_not_null
                    / (
                        temperature_features.temperature_not_null
                        + temperature_features.temperature_null
                    )
                ) <= 0.5
                # check against median in case start/end of data does not cover a full period
                invalid_temperature_rows |= (
                    temperature_features.temperature_not_null <= median_samples * 0.5
                )

                if invalid_temperature_rows.any():
                    self.warnings.append(
                        EEMeterWarning(
                            qualified_name="eemeter.sufficiency_criteria.missing_high_frequency_temperature_data",
                            description=(
                                "More than 50% of the high frequency temperature data is missing."
                            ),
                            data=[
                                timestamp.isoformat()
                                for timestamp in invalid_temperature_rows.index
                            ],
                        )
                    )
                    temperature_features.loc[
                        invalid_temperature_rows, "temperature_mean"
                    ] = np.nan

        temp = temperature_features["temperature_mean"].rename("temperature")
        features = temperature_features.drop(columns=["temperature_mean"])
        return temp, features

    # TODO: DELETE THIS after making real billing data class
    @property
    def billing_df(self) -> pd.DataFrame | None:
        """Get the corrected input data stored in the class. The actual dataframe is immutable, this returns a copy."""

        df = self._df.copy()

        # find indices where observed changes from prior
        observed_change = df["observed"].diff()
        observed_change = observed_change[observed_change != 0].index
        obs_change_idx = df.index.get_indexer(observed_change)
        obs_change_idx = np.append(obs_change_idx, len(df))
        obs_change_idx = np.delete(obs_change_idx, np.where(np.diff(obs_change_idx) < 15)[0])

        if obs_change_idx[0] != 0:
            obs_change_idx = np.insert(obs_change_idx, 0, 0)

        # create vector where value increases at each observed change
        group = []
        for i in range(1, len(obs_change_idx)):
            idx_range = obs_change_idx[i] - obs_change_idx[i-1]

            group.extend([i] * idx_range)

        df["group"] = group

        # get median delta

        # get first datetime, average temperature, sum of observed for each group and make new df
        df_temp = df.reset_index()
        df_temp = df_temp.rename(columns={"index": "datetime"})

        df_grouped = df_temp.groupby("group").agg({
            "datetime": "first",
            "season": "first",
            "weekday_weekend": "first",
            "temperature": "mean",
            "observed": "mean",
        }).set_index("datetime")

        # create days column for number of days between current and previous index
        df_grouped["days"] = df_grouped.index.to_series().diff().dt.days

        df_grouped = df_grouped.dropna()

        # create weights from days column
        df_grouped["weights"] = df_grouped["days"] / df_grouped["days"].sum()

        df_grouped = df_grouped.drop(columns=["days"])

        if self._df is None:
            return None
        else:
            return df_grouped.copy()


class BillingBaselineData(_BillingData):
    """
    Data class to represent Billing Baseline Data.

    Only baseline data should go into the dataframe input, no blackout data should be input.
    Checks sufficiency for the data provided as input depending on OpenEEMeter specifications and populates disqualifications and warnings based on it.

    Billing data should have an extra month's data appended at the to denote end of period. (Do not append NaN, any other value would work.)

    Args:
        df (DataFrame): A dataframe having a datetime index or a datetime column with the timezone also being set.
            It also requires 2 more columns - 'observed' for meter data, and 'temperature' for temperature data.
            The temperature column should have values in Fahrenheit. Please convert your temperatures accordingly.

        is_electricity_data (bool): Flag to ascertain if this is electricity data or not. Electricity data values of 0 are set to NaN.

    Attributes:
        df (DataFrame): Immutable dataframe that contains the meter and temperature values for the baseline data period.
        disqualification (list[EEMeterWarning]): A list of serious issues with the data that can degrade the quality of the model. If you want to go ahead with building the model while ignoring them, set the ignore_disqualification = True flag in the model. By default disqualifications are not ignored.
        warnings (list[EEMeterWarning]): A list of ssues with the data, but none that will severely reduce the quality of the model built.
    """

    def _check_data_sufficiency(self, sufficiency_df):
        """
        Private method which checks the sufficiency of the data for billing baseline calculations using the predefined OpenEEMeter sufficiency criteria.

        Args:
            sufficiency_df (pandas.DataFrame): DataFrame containing the data for sufficiency check. Should have features such as -
            temperature_null: number of temperature null periods in each aggregation step
            temperature_not_null: number of temperature non null periods in each aggregation step

        Returns:
            disqualification (List): List of disqualifications
            warnings (list): List of warnings

        """
        bsc = BillingSufficiencyCriteria(
            data=sufficiency_df, is_electricity_data=self.is_electricity_data
        )
        bsc.check_sufficiency_baseline()
        disqualification = bsc.disqualification
        warnings = bsc.warnings

        # _, disqualification, warnings = sufficiency_criteria_baseline(
        #     sufficiency_df,
        #     is_reporting_data=False,
        #     is_electricity_data=self.is_electricity_data,
        # )
        return disqualification, warnings


class BillingReportingData(_BillingData):
    """Data class to represent Billing Reporting Data.

    Only reporting data should go into the dataframe input, no blackout data should be input.
    Checks sufficiency for the data provided as input depending on OpenEEMeter specifications and populates disqualifications and warnings based on it.

    Meter data input is optional for the reporting class.

    Args:
        df (DataFrame): A dataframe having a datetime index or a datetime column with the timezone also being set.
            It also requires 2 more columns - 'observed' for meter data, and 'temperature' for temperature data.
            The temperature column should have values in Fahrenheit. Please convert your temperatures accordingly.

        is_electricity_data (bool): Flag to ascertain if this is electricity data or not. Electricity data values of 0 are set to NaN.

    Attributes:
        df (DataFrame): Immutable dataframe that contains the meter and temperature values for the baseline data period.
        disqualification (list[EEMeterWarning]): A list of serious issues with the data that can degrade the quality of the model. If you want to go ahead with building the model while ignoring them, set the ignore_disqualification = True flag in the model. By default disqualifications are not ignored.
        warnings (list[EEMeterWarning]): A list of ssues with the data, but none that will severely reduce the quality of the model built.
    """

    def __init__(self, df: pd.DataFrame, is_electricity_data: bool):
        df = df.copy()
        if "observed" not in df.columns:
            df["observed"] = np.nan

        super().__init__(df, is_electricity_data)

    @classmethod
    def from_series(
        cls,
        meter_data: pd.Series | pd.DataFrame | None,
        temperature_data: pd.Series | pd.DataFrame,
        is_electricity_data: bool,
        tzinfo: datetime.tzinfo | None = None,
    ):
        """Create a BillingReportingData instance from meter data and temperature data.

        Args:
            meter_data: The meter data to be used for the BillingReportingData instance.
            temperature_data: The temperature data to be used for the BillingReportingData instance.
            is_electricity_data: Flag indicating whether the meter data represents electricity data.
            tzinfo: Timezone information to be used for the meter data.

        Returns:
            An instance of the Data class.
        """
        if tzinfo and meter_data is not None:
            raise ValueError(
                "When passing meter data to BillingReportingData, convert its DatetimeIndex to local timezone first; `tzinfo` param should only be used in the absence of reporting meter data."
            )
        if is_electricity_data is None and meter_data is not None:
            raise ValueError(
                "Must specify is_electricity_data when passing meter data."
            )
        if meter_data is None:
            meter_data = pd.DataFrame(
                {"observed": np.nan}, index=temperature_data.index
            )
            if tzinfo:
                meter_data = meter_data.tz_convert(tzinfo)
        if meter_data.empty:
            raise ValueError(
                "Pass meter_data=None to explicitly create a temperature-only reporting data instance."
            )
        return super().from_series(meter_data, temperature_data, is_electricity_data)

    def _check_data_sufficiency(self, sufficiency_df):
        """
        Private method which checks the sufficiency of the data for billing reporting calculations using the predefined OpenEEMeter sufficiency criteria.

        Parameters
        ----------
        1. sufficiency_df (pandas.DataFrame): DataFrame containing the data for sufficiency check. Should have features such as -
            - temperature_null: number of temperature null periods in each aggregation step
            - temperature_not_null: number of temperature non null periods in each aggregation step

        Returns
        -------
            disqualification (List): List of disqualifications
            warnings (list): List of warnings

        """
        bsc = BillingSufficiencyCriteria(data=sufficiency_df, is_reporting_data=True)
        bsc.check_sufficiency_reporting()
        disqualification = bsc.disqualification
        warnings = bsc.warnings

        # _, disqualification, warnings = sufficiency_criteria_baseline(
        #     sufficiency_df,
        #     is_reporting_data=True,
        #     is_electricity_data=self.is_electricity_data,
        # )
        return disqualification, warnings
